__author__ = "Simon Nilsson"

import functools
import multiprocessing
import os
import platform
from typing import Any, Dict, Optional, Tuple, Union

import cv2
import numpy as np
import pandas as pd

from simba.data_processors.directing_other_animals_calculator import \
    DirectingOtherAnimalsAnalyzer
from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_keys_exist_in_dict,
                                check_if_valid_rgb_tuple, check_int,
                                check_valid_lst,
                                check_video_and_data_frm_count_align)
from simba.utils.data import create_color_palettes
from simba.utils.enums import Formats, TextOptions
from simba.utils.errors import AnimalNumberError, NoFilesFoundError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder,
                                    find_core_cnt, get_fn_ext,
                                    get_video_meta_data, read_df)
from simba.utils.warnings import NoDataFoundWarning

DIRECTION_THICKNESS = "direction_thickness"
DIRECTIONALITY_COLOR = "directionality_color"
CIRCLE_SIZE = "circle_size"
HIGHLIGHT_ENDPOINTS = "highlight_endpoints"
SHOW_POSE = "show_pose"
ANIMAL_NAMES = "animal_names"
STYLE_ATTR = [
    DIRECTION_THICKNESS,
    DIRECTIONALITY_COLOR,
    CIRCLE_SIZE,
    HIGHLIGHT_ENDPOINTS,
    SHOW_POSE,
    ANIMAL_NAMES,
]


def _directing_animals_mp(
    frm_range: Tuple[int, np.ndarray],
    directionality_data: pd.DataFrame,
    pose_data: pd.DataFrame,
    style_attr: dict,
    animal_bp_dict: dict,
    save_temp_dir: str,
    video_path: str,
    video_meta_data: dict,
    colors: list,
):
    batch = frm_range[0]
    fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
    start_frm, current_frm, end_frm = frm_range[1][0], frm_range[1][0], frm_range[1][-1]
    save_path = os.path.join(save_temp_dir, f"{batch}.mp4")
    writer = cv2.VideoWriter(
        save_path,
        fourcc,
        video_meta_data["fps"],
        (video_meta_data["width"], video_meta_data["height"]),
    )
    cap = cv2.VideoCapture(video_path)
    cap.set(1, start_frm)
    while current_frm <= end_frm:
        ret, img = cap.read()
        if ret:
            frm_data = pose_data.iloc[current_frm]
            if style_attr[SHOW_POSE]:
                for animal_cnt, (animal_name, animal_bps) in enumerate(
                    animal_bp_dict.items()
                ):
                    for bp_cnt, bp in enumerate(
                        zip(animal_bps["X_bps"], animal_bps["Y_bps"])
                    ):
                        x_bp, y_bp = frm_data[bp[0]], frm_data[bp[1]]
                        cv2.circle(
                            img,
                            (int(x_bp), int(y_bp)),
                            style_attr[CIRCLE_SIZE],
                            animal_bp_dict[animal_name]["colors"][bp_cnt],
                            -1,
                        )
            if style_attr[ANIMAL_NAMES]:
                for animal_name, bp_data in animal_bp_dict.items():
                    headers = [bp_data["X_bps"][-1], bp_data["Y_bps"][-1]]
                    bp_cords = pose_data.loc[current_frm, headers].values.astype(
                        np.int64
                    )
                    cv2.putText(
                        img,
                        animal_name,
                        (bp_cords[0], bp_cords[1]),
                        TextOptions.FONT.value,
                        2,
                        animal_bp_dict[animal_name]["colors"][0],
                        1,
                    )

            if current_frm in list(directionality_data["Frame_#"].unique()):
                img_data = directionality_data[
                    directionality_data["Frame_#"] == current_frm
                ]
                for animal_name in img_data["Animal_1"].unique():
                    animal_img_data = img_data[
                        img_data["Animal_1"] == animal_name
                    ].reset_index(drop=True)
                    img = PlottingMixin.draw_lines_on_img(
                        img=img,
                        start_positions=animal_img_data[
                            ["Eye_x", "Eye_y"]
                        ].values.astype(np.int64),
                        end_positions=animal_img_data[
                            ["Animal_2_bodypart_x", "Animal_2_bodypart_y"]
                        ].values.astype(np.int64),
                        color=tuple(colors[animal_name]),
                        highlight_endpoint=style_attr[HIGHLIGHT_ENDPOINTS],
                        thickness=style_attr[DIRECTION_THICKNESS],
                        circle_size=style_attr[CIRCLE_SIZE],
                    )
            current_frm += 1
            writer.write(np.uint8(img))
            print(
                f"Frame: {current_frm} / {video_meta_data['frame_count']}. Core batch: {batch}"
            )

        else:
            break

    writer.release()
    return batch


class DirectingOtherAnimalsVisualizerMultiprocess(ConfigReader, PlottingMixin):
    """
    Class for visualizing when animals are directing towards body-parts of other animals using multiprocessing.

    .. important::
       Requires the pose-estimation data for the left ear, right ears and nose of individual animals.

    .. note::
        Example of expected output https://www.youtube.com/watch?v=d6pAatreb1E&list=PLi5Vwf0hhy1R6NDQJ3U28MOUJPfl2YWYl&index=22

       `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#visualizing-data-tables`>__.

    .. image:: _static/img/directing_other_animals.png
       :width: 450
       :align: center

    :param Union[str, os.PathLike] config_path: path to SimBA project config file in Configparser format
    :param Union[str, os.PathLike] video_path: Path to video for to visualize directionality.
    :param Dict[str, Any] style_attr: Video style attribitions (colors and sizes etc.)
    :param Optional[int] core_cnt: How many cores to use to create the video. Deafult -1 which is all the cores.

    :examples:
    >>> style_attr = {'show_pose': True, 'animal_names': False, 'circle_size': 3, 'directionality_color': [(255, 0, 0), (0, 0, 255)], 'direction_thickness': 10, 'highlight_endpoints': True}
    >>> test = DirectingOtherAnimalsVisualizerMultiprocess(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini', video_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/videos/Together_1.avi', style_attr=style_attr, core_cnt=-1)
    >>> test.run()
    """

    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        video_path: Union[str, os.PathLike],
        style_attr: Dict[str, Any],
        core_cnt: Optional[int] = -1,
    ):

        if platform.system() == "Darwin":
            multiprocessing.set_start_method("spawn", force=True)
        check_file_exist_and_readable(file_path=video_path)
        check_file_exist_and_readable(file_path=config_path)
        check_if_keys_exist_in_dict(
            data=style_attr,
            key=STYLE_ATTR,
            name=f"{self.__class__.__name__} style_attr",
        )
        check_int(
            name=f"{self.__class__.__name__} core_cnt",
            value=core_cnt,
            min_value=-1,
            max_value=find_core_cnt()[0],
        )
        if core_cnt == -1:
            core_cnt = find_core_cnt()[0]
        ConfigReader.__init__(self, config_path=config_path)
        PlottingMixin.__init__(self)
        if self.animal_cnt < 2:
            raise AnimalNumberError(
                "Cannot analyze directionality between animals in a project with less than two animals.",
                source=self.__class__.__name__,
            )
        self.animal_names = [k for k in self.animal_bp_dict.keys()]
        _, self.video_name, _ = get_fn_ext(video_path)
        self.data_path = os.path.join(
            self.outlier_corrected_dir, f"{self.video_name}.{self.file_type}"
        )
        if not os.path.isfile(self.data_path):
            raise NoFilesFoundError(
                msg=f"SIMBA ERROR: Could not find the file at path {self.data_path}. Make sure the data file exist to create directionality visualizations",
                source=self.__class__.__name__,
            )
        self.direction_analyzer = DirectingOtherAnimalsAnalyzer(
            config_path=config_path,
            bool_tables=False,
            summary_tables=False,
            aggregate_statistics_tables=False,
            data_paths=self.data_path,
        )
        self.direction_analyzer.run()
        self.direction_analyzer.transpose_results()
        self.fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        self.style_attr = style_attr
        self.direction_colors = {}
        if isinstance(self.style_attr[DIRECTIONALITY_COLOR], list):
            check_valid_lst(
                data=self.style_attr[DIRECTIONALITY_COLOR],
                source=f"{self.__class__.__name__} colors",
                valid_dtypes=(tuple,),
                min_len=self.animal_cnt,
            )
            for i in range(len(self.animal_names)):
                check_if_valid_rgb_tuple(data=self.style_attr[DIRECTIONALITY_COLOR][i])
                self.direction_colors[self.animal_names[i]] = self.style_attr[
                    DIRECTIONALITY_COLOR
                ][i]
        if isinstance(self.style_attr[DIRECTIONALITY_COLOR], tuple):
            check_if_valid_rgb_tuple(self.style_attr[DIRECTIONALITY_COLOR])
            for i in range(len(self.animal_names)):
                self.direction_colors[self.animal_names[i]] = self.style_attr[
                    DIRECTIONALITY_COLOR
                ]
        else:
            self.random_colors = create_color_palettes(1, int(self.animal_cnt**2))[0]
            self.random_colors = [
                [int(item) for item in sublist] for sublist in self.random_colors
            ]
            for cnt in range(len(self.animal_names)):
                self.direction_colors[self.animal_names[cnt]] = self.random_colors[cnt]
        self.data_dict = self.direction_analyzer.directionality_df_dict
        if not os.path.exists(self.directing_animals_video_output_path):
            os.makedirs(self.directing_animals_video_output_path)
        self.data_df = read_df(self.data_path, file_type=self.file_type)
        self.video_save_path = os.path.join(
            self.directing_animals_video_output_path, f"{self.video_name}.mp4"
        )
        self.cap = cv2.VideoCapture(video_path)
        self.video_meta_data = get_video_meta_data(video_path)
        check_video_and_data_frm_count_align(
            video=video_path, data=self.data_path, name=video_path, raise_error=False
        )
        self.video_save_path = os.path.join(
            self.directing_animals_video_output_path, f"{self.video_name}.mp4"
        )
        if not os.path.exists(self.directing_animals_video_output_path):
            os.makedirs(self.directing_animals_video_output_path)
        self.save_path = os.path.join(
            self.directing_animals_video_output_path, self.video_name + ".mp4"
        )
        self.save_temp_path = os.path.join(
            self.directing_animals_video_output_path, "temp"
        )
        if os.path.exists(self.save_temp_path):
            self.remove_a_folder(folder_dir=self.save_temp_path)
        os.makedirs(self.save_temp_path)
        self.core_cnt, self.video_path = core_cnt, video_path
        print(f"Processing video {self.video_name}...")

    def run(self):
        video_data = self.data_dict[self.video_name]
        self.writer = cv2.VideoWriter(
            self.video_save_path,
            self.fourcc,
            self.video_meta_data["fps"],
            (self.video_meta_data["width"], self.video_meta_data["height"]),
        )
        if len(video_data) < 1:
            NoDataFoundWarning(
                msg=f"SimBA skipping video {self.video_name}: No animals are directing each other in the video."
            )
        else:
            frm_data = np.array_split(
                list(range(0, self.video_meta_data["frame_count"] + 1)), self.core_cnt
            )
            frm_ranges = []
            for i in range(len(frm_data)):
                frm_ranges.append((i, frm_data[i]))
            print(
                f"Creating directing images, multiprocessing (chunksize: {self.multiprocess_chunksize}, cores: {self.core_cnt})..."
            )
            with multiprocessing.Pool(
                self.core_cnt, maxtasksperchild=self.maxtasksperchild
            ) as pool:
                constants = functools.partial(
                    _directing_animals_mp,
                    directionality_data=video_data,
                    pose_data=self.data_df,
                    video_meta_data=self.video_meta_data,
                    style_attr=self.style_attr,
                    save_temp_dir=self.save_temp_path,
                    video_path=self.video_path,
                    animal_bp_dict=self.animal_bp_dict,
                    colors=self.direction_colors,
                )
                for cnt, result in enumerate(
                    pool.imap(
                        constants, frm_ranges, chunksize=self.multiprocess_chunksize
                    )
                ):
                    print(f"Core batch {result+1} complete...")
            print(f"Joining {self.video_name} multi-processed video...")
            concatenate_videos_in_folder(
                in_folder=self.save_temp_path,
                save_path=self.save_path,
                video_format="mp4",
                remove_splits=True,
            )
            self.timer.stop_timer()
            pool.terminate()
            pool.join()
            stdout_success(
                msg=f"Video {self.video_name} complete. Video saved in {self.directing_animals_video_output_path} directory",
                elapsed_time=self.timer.elapsed_time_str,
            )


# style_attr = {SHOW_POSE: True,
#               ANIMAL_NAMES: False,
#               CIRCLE_SIZE: 3,
#               DIRECTIONALITY_COLOR: [(255, 0, 0), (0, 0, 255)],
#               DIRECTION_THICKNESS: 10,
#               HIGHLIGHT_ENDPOINTS: True}
# test = DirectingOtherAnimalsVisualizerMultiprocess(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                                    video_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/videos/Together_1.avi',
#                                                    style_attr=style_attr,
#                                                    core_cnt=-1)
#
# test.run()


# style_attr = {'Show_pose': True,
#               'Pose_circle_size': 3,
#               "Direction_color": 'Random',
#               'Direction_thickness': 4,
#               'Highlight_endpoints': True,
#               'Polyfill': True}
# test = DirectingOtherAnimalsVisualizerMultiprocess(config_path='/Users/simon/Desktop/envs/troubleshooting/sleap_5_animals/project_folder/project_config.ini',
#                                                    data_path='/Users/simon/Desktop/envs/troubleshooting/sleap_5_animals/project_folder/csv/outlier_corrected_movement_location/Testing_Video_3.csv',
#                                                    style_attr=style_attr,
#                                                    core_cnt=5)
#
# test.run()


# style_attr = {'Show_pose': True, 'Pose_circle_size': 3, "Direction_color": 'Random', 'Direction_thickness': 4, 'Highlight_endpoints': True, 'Polyfill': True}
# test = DirectingOtherAnimalsVisualizerMultiprocess(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                        data_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/outlier_corrected_movement_location/Together_1.csv',
#                                        style_attr=style_attr,
#                                                    core_cnt=5)
#
# test.run()



__author__ = "Simon Nilsson"

import os
from itertools import permutations
from typing import Dict, Optional, Tuple, Union

import numpy as np
import pandas as pd

try:
    from typing import Literal
except:
    from typing_extensions import Literal

from shapely.geometry import MultiPoint

from simba.mixins.config_reader import ConfigReader
from simba.mixins.geometry_mixin import GeometryMixin
from simba.mixins.unsupervised_mixin import UnsupervisedMixin
from simba.unsupervised.enums import Clustering, Unsupervised
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_if_keys_exist_in_dict,
                                check_if_valid_rgb_tuple, check_int, check_str)
from simba.utils.data import sample_df_n_by_unique
from simba.utils.read_write import read_pickle


class ClusterVideoVisualizer(ConfigReader, UnsupervisedMixin):
    """
    Class for creating video examples of cluster assignments.

    :param Union[str, os.PathLike] config_path: Path to SimBA project configuration file.
    :param Union[str, os.PathLike] data_path: Path to pickle file containing unsupervised results.
    :param Optional[Union[int, None]] max_videos: Maximum number of videos to create for each cluster. Defaults to None.
    :param Optional[int] speed: Speed of the generated videos. Defaults to 1.0.
    :param Optional[Tuple[int, int, int]] bg_clr: Background color of the videos as RGB tuple. Defaults to white (255, 255, 255).
    :param Optional[Literal] plot_type: Type of plot to generate ('VIDEO', 'HULL', 'SKELETON', 'POINTS'). Defaults to 'SKELETON'.

    :example:
    >>> config_path = '/Users/simon/Desktop/envs/simba/troubleshooting/NG_Unsupervised/project_folder/project_config.ini'
    >>> data_path = '/Users/simon/Desktop/envs/simba/troubleshooting/NG_Unsupervised/project_folder/cluster_mdls/hopeful_khorana.pickle'
    >>> visualizer = ClusterVideoVisualizer(config_path=config_path, data_path=data_path, bg_clr=(0, 0, 255), max_videos=20, speed=0.2, plot_type='POINTS')
    >>> visualizer.run()
    """

    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        data_path: Union[str, os.PathLike],
        max_videos: Optional[Union[int, None]] = None,
        speed: Optional[int] = 1.0,
        bg_clr: Optional[Tuple[int, int, int]] = (255, 255, 255),
        plot_type: Optional[
            Literal["VIDEO", "HULL", "SKELETON", "POINTS"]
        ] = "SKELETON",
    ):

        check_file_exist_and_readable(file_path=data_path)
        check_file_exist_and_readable(file_path=config_path)
        check_if_valid_rgb_tuple(data=bg_clr)
        if max_videos != None:
            check_int(name="max_videos", value=max_videos, min_value=1)
        check_float(name="speed", value=speed, min_value=0.1)
        check_str(
            name="plot_type",
            value=plot_type,
            options=("VIDEO", "HULL", "SKELETON", "POINTS"),
        )
        ConfigReader.__init__(self, config_path=config_path)
        UnsupervisedMixin.__init__(self)
        self.data = read_pickle(data_path=data_path)
        check_if_keys_exist_in_dict(
            data=self.data,
            key=[Clustering.CLUSTER_MODEL.value, Unsupervised.DATA.value],
            name=data_path,
        )
        self.max_videos, self.speed, self.plot_type, self.bg_clr = (
            max_videos,
            speed,
            plot_type,
            bg_clr,
        )
        self.cl_mdl_name = self.data[Clustering.CLUSTER_MODEL.value][
            Unsupervised.HASHED_NAME.value
        ]
        self.animal_bp_cols, self.skeleton_perm = {}, {}
        self.save_dir = os.path.join(
            self.project_path, self.frames_output_dir, "clusters"
        )
        if not os.path.isdir(self.save_dir):
            os.makedirs(self.save_dir)
        for animal_nme in self.animal_bp_dict.keys():
            animal_bp_cols = []
            for x, y in zip(
                self.animal_bp_dict[animal_nme]["X_bps"],
                self.animal_bp_dict[animal_nme]["Y_bps"],
            ):
                animal_bp_cols.extend((x, y))
            self.animal_bp_cols[animal_nme] = animal_bp_cols
            bp_names = [x[:-2] for x in self.animal_bp_dict[animal_nme]["X_bps"]]
            self.skeleton_perm[animal_nme] = list(permutations(bp_names, 2))

    def run(self):
        cluster_idx = self.data[Unsupervised.DATA.value][
            Unsupervised.BOUTS_FEATURES.value
        ].index
        cluster_lbls = self.data[Clustering.CLUSTER_MODEL.value][
            Unsupervised.MODEL.value
        ].labels_
        cluster_df = pd.DataFrame(cluster_lbls, columns=["CLUSTER"], index=cluster_idx)
        if self.max_videos != None:
            cluster_df = sample_df_n_by_unique(
                df=cluster_df, field="CLUSTER", n=self.max_videos
            )
            cluster_df = cluster_df[~cluster_df.index.duplicated(keep="first")]
        for cluster_id in sorted(cluster_df["CLUSTER"].unique()):
            event_idx = list(cluster_df[cluster_df["CLUSTER"] == cluster_id].index)
            print(f"Creating {len(event_idx)} videos for cluster {cluster_id}...")
            for event in event_idx:
                video_pose = (
                    self.data[Unsupervised.DATA.value][Unsupervised.FRAME_POSE.value]
                    .loc[event[0], :]
                    .reset_index()
                )
                event_df = video_pose[
                    (video_pose["FRAME"] >= event[1])
                    & (video_pose["FRAME"] <= event[2])
                ].astype(np.int64)
                save_path = os.path.join(
                    self.save_dir, f"{event[0]}_{event[1]}_{event[2]}_{cluster_id}.mp4"
                )
                video_info, _, fps = self.read_video_info(video_name=event[0])
                out_fps = int(fps * self.speed)
                if out_fps > 1:
                    out_fps = 1
                w, h = int(video_info["Resolution_width"]), int(
                    video_info["Resolution_height"].astype(int)
                )
                bg_img = np.full((w, h, 3), self.bg_clr, dtype=np.uint8)
                shapes = []
                for name, bps in self.animal_bp_cols.items():
                    if self.plot_type == "HULL":
                        animal_event_pose_data = event_df[bps].values
                        animal_event_pose_data = animal_event_pose_data.reshape(
                            len(animal_event_pose_data), -1, 2
                        )
                        shapes.append(
                            GeometryMixin().multiframe_bodyparts_to_polygon(
                                data=animal_event_pose_data,
                                pixels_per_mm=1,
                                parallel_offset=1,
                                verbose=False,
                                core_cnt=-1,
                            )
                        )
                    elif self.plot_type == "SKELETON":
                        animal_skeleton_bps = self.skeleton_perm[name]
                        shapes.append(
                            GeometryMixin().multiframe_bodyparts_to_multistring_skeleton(
                                data_df=event_df,
                                skeleton=animal_skeleton_bps,
                                core_cnt=-1,
                                verbose=False,
                            )
                        )
                    elif self.plot_type == "POINTS":
                        animal_event_pose_data = event_df[bps].values
                        animal_event_pose_data = animal_event_pose_data.reshape(
                            len(animal_event_pose_data), -1, 2
                        )
                        results = GeometryMixin().multiframe_bodypart_to_point(
                            data=animal_event_pose_data
                        )
                        multi_points = []
                        for frm in range(len(results)):
                            multi_points.append(MultiPoint(results[frm]))
                        shapes.append(multi_points)
                _ = GeometryMixin.geometry_video(
                    shapes=shapes,
                    save_path=save_path,
                    fps=out_fps,
                    size=(w, h),
                    bg_img=bg_img,
                )
                print(f"Cluster video saved at {save_path}...")


# config_path = '/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/project_config.ini'
# data_path = '/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/small_clusters/adoring_hoover.pickle'
# visualizer = ClusterVideoVisualizer(config_path=config_path,
#                                     data_path=data_path,
#                                     bg_clr=(0, 0, 255),
#                                     max_videos=3,
#                                     speed=1.0,
#                                     plot_type='HULL')
# visualizer.run()



__author__ = "Simon Nilsson"

import os
from copy import deepcopy
from typing import List, Optional, Union

import pandas as pd

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import (check_file_exist_and_readable, check_int,
                                check_str, check_valid_lst)
from simba.utils.data import df_smoother, savgol_smoother
from simba.utils.enums import TagNames
from simba.utils.errors import InvalidInputError, NoFilesFoundError
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import (copy_files_to_directory,
                                    find_files_of_filetypes_in_directory,
                                    find_video_of_file, get_fn_ext,
                                    get_video_meta_data, read_df,
                                    read_video_info, write_df)


class Smoothing(ConfigReader):
    """
    Smooth pose-estimation data according to user-defined method.

    .. image:: _static/img/smoothing.gif
       :width: 600
       :align: center

    .. video:: _static/img/smoothing_example_2.webm
       :width: 800
       :autoplay:
       :loop:

    .. note::
       `Smoothing tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Scenario1.md#to-import-multiple-dlc-csv-files>`__.

    .. important::
        The smoothened data overwrites the original data on disk. If the original data is required, pass ``copy_originals = True`` to save a copy of the original data.

    :param Union[str, os.PathLike] config_path: path to SimBA project config file in Configparser format.
    :param Union[str, os.PathLike, List[Union[str, os.PathLike]]] data_path: Path to directory containing pose-estimation data, to a file containing pose-estimation data, or a list of paths containing pose-estimation data.
    :param int time_window: Rolling time window in millisecond to use when smoothing. Larger time-windows and greater smoothing.
    :param Optional[Literal["gaussian", "savitzky-golay"]] method: Type of smoothing_method. OPTIONS: ``gaussian``, ``savitzky-golay``. Default `gaussian`.
    :param bool multi_index_df_headers: If True, the incoming data is multi-index columns dataframes. Default: False.
    :param bool copy_originals: If truth-like, then the pre-smoothened, original data, will be bo stored in a subdirectory of the original data. The subdirectory is named according to the type of smoothing method and datetime of the operation.

    :references:
        .. [1] `Video expected putput <https://www.youtube.com/watch?v=d9-Bi4_HyfQ>`__.

    :examples:
    >>> smoother = Smoothing(data_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/csv/input_csv/Together_1.csv', config_path=r'/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini', method='Savitzky-Golay', time_window=500, multi_index_df_headers=True, copy_originals=True)
    >>> smoother.run()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 data_path: Union[str, os.PathLike, List[Union[str, os.PathLike]]],
                 time_window: int,
                 method: Optional[Literal["gaussian", "savitzky-golay"]] = 'Savitzky-Golay',
                 multi_index_df_headers: Optional[bool] = False,
                 copy_originals: Optional[bool] = False) -> None:

        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        log_event(logger_name=str(self.__class__.__name__), log_type=TagNames.CLASS_INIT.value, msg=self.create_log_msg_from_init_args(locals=locals()))
        if isinstance(data_path, list):
            check_valid_lst(data=data_path, source=self.__class__.__name__, valid_dtypes=(str,))
            for i in data_path: check_file_exist_and_readable(file_path=i)
            self.file_paths = deepcopy(data_path)
        elif os.path.isdir(data_path):
            self.file_paths = find_files_of_filetypes_in_directory(directory=data_path, extensions=[f'.{self.file_type}'], raise_error=True)
        elif os.path.isfile(data_path):
            check_file_exist_and_readable(file_path=data_path)
            self.file_paths = [data_path]
        else:
            raise InvalidInputError(msg=f'{data_path} is not a valid data directory, or a valid file path, or a valid list of file paths', source=self.__class__.__name__)
        check_int(value=time_window, min_value=1, name=f'{self.__class__.__name__} time_window')
        check_str(name=f'{self.__class__.__name__} method', value=method.lower(), options=("gaussian", "savitzky-golay"))
        if copy_originals:
            self.originals_dir = os.path.join(os.path.dirname(self.file_paths[0]), f"Pre_{method}_{time_window}_smoothing_{self.datetime}")
            os.makedirs(self.originals_dir)
        self.multi_index_df_headers, self.method, self.time_window, self.copy_originals = multi_index_df_headers, method.lower(), time_window, copy_originals

    def __insert_multiindex_header(self, df: pd.DataFrame):
        multi_idx_header = []
        for i in range(len(df.columns)):
            multi_idx_header.append(("IMPORTED_POSE", "IMPORTED_POSE", list(df.columns)[i]))
        df.columns = pd.MultiIndex.from_tuples(multi_idx_header)
        return df

    def run(self):
        print(f'Running smoothing on {len(self.file_paths)} data file(s)...')
        for file_cnt, file_path in enumerate(self.file_paths):
            df = read_df(file_path=file_path, file_type=self.file_type, check_multiindex=True)
            video_timer = SimbaTimer(start=True)
            _, video_name, _ = get_fn_ext(filepath=file_path)
            video_path = find_video_of_file(video_dir=self.video_dir, filename=video_name, raise_error=False, warning=False)
            if video_path is None:
                if not os.path.isfile(self.video_info_path):
                    raise NoFilesFoundError(msg=f"To perform smoothing, SimBA needs to read the video FPS. SimBA could not find the video {video_name} in represented in the {self.video_dir} directory or in {self.video_info_path} file. Please import the video and/or include it in the video_logs.csv file so SimBA can know the video FPS", source=self.__class__.__name__)
                else:
                    self.video_info_df = self.read_video_info_csv(file_path=self.video_info_path)
                    video_info = read_video_info(vid_info_df=self.video_info_df,video_name=video_name, raise_error=False)
                    if video_info is None:
                        raise NoFilesFoundError(msg=f"To perform smoothing, SimBA needs to read the video FPS. SimBA could not find the video {video_name} in represented in the {self.video_dir} directory or in {self.video_info_path} file. Please import the video and/or include it in the video_logs.csv file so SimBA can know the video FPS", source=self.__class__.__name__)
                    fps = video_info[2]
            else:
                fps = get_video_meta_data(video_path=video_path)['fps']
            if self.method == 'savitzky-golay':
                df = savgol_smoother(data=df, fps=fps, time_window=self.time_window, source=video_name)
            else:
                df = df_smoother(data=df, fps=fps, time_window=self.time_window, source=video_name, method='gaussian')
            if self.multi_index_df_headers:
                df = self.__insert_multiindex_header(df=df)
            if self.copy_originals:
                copy_files_to_directory(file_paths=[file_path], dir=self.originals_dir)
            write_df(df=df, file_type=self.file_type, save_path=file_path, multi_idx_header=self.multi_index_df_headers)
            video_timer.stop_timer()
            print(f"Video {video_name} smoothed ({self.method}: {str(self.time_window)}ms) (elapsed time {video_timer.elapsed_time_str})...")
        self.timer.stop_timer()
        if self.copy_originals:
            msg = f"{len(self.file_paths)} data file(s) smoothened using {self.method} method and {self.time_window} time-window. Originals saved in {self.originals_dir} directory."
        else:
            msg = f"{len(self.file_paths)} data file(s) smoothened using {self.method} method and {self.time_window} time-window."
        stdout_success(msg=msg, elapsed_time=self.timer.elapsed_time_str, source=self.__class__.__name__)


__author__ = "Simon Nilsson"

import os
import threading
from tkinter import *
from typing import Union

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import (DropDownMenu, Entry_Box, FileSelect,
                                        FolderSelect)
from simba.unsupervised.enums import UMLOptions, Unsupervised
from simba.unsupervised.tsne import TSNEGridSearch
from simba.unsupervised.umap_embedder import UmapEmbedder
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists)
from simba.utils.enums import Formats, Options
from simba.utils.errors import NoSpecifiedOutputError


class FitDimReductionPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: Union[str, os.PathLike]):
        PopUpMixin.__init__(self, title="FIT DIMENSIONALITY REDUCTION MODELS")
        ConfigReader.__init__(self, config_path=config_path)
        self.variance_options = [
            str(x) + "%" for x in UMLOptions.VARIANCE_OPTIONS.value
        ]
        self.dataset_frm = LabelFrame(
            self.main_frm,
            text="DATASET",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.dataset_file_selected = FileSelect(
            self.dataset_frm,
            "DATASET (PICKLE):",
            lblwidth=25,
            file_types=[("Dataset pickle", "*.pickle")],
            initialdir=self.logs_path,
        )
        self.dataset_frm.grid(row=0, column=0, sticky=NW)
        self.dataset_file_selected.grid(row=0, column=0, sticky=NW)

        self.save_frm = LabelFrame(
            self.main_frm,
            text="SAVE",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.save_dir = FolderSelect(
            self.save_frm, "SAVE DIRECTORY:", lblwidth=25, initialdir=self.project_path
        )
        self.save_frm.grid(row=1, column=0, sticky=NW)
        self.save_dir.grid(row=0, column=0, sticky=NW)

        settings_frm = LabelFrame(
            self.main_frm,
            text="SETTINGS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.scaling_dropdown = DropDownMenu(
            settings_frm, "SCALING:", Options.SCALER_NAMES.value, "25"
        )
        self.scaling_dropdown.setChoices(Options.SCALER_NAMES.value[0])
        self.var_threshold_dropdown = DropDownMenu(
            settings_frm, "VARIANCE THRESHOLD:", self.variance_options, "25"
        )
        self.var_threshold_dropdown.setChoices(self.variance_options[0])
        self.algo_dropdown = DropDownMenu(
            settings_frm,
            "ALGORITHM:",
            UMLOptions.DR_ALGO_OPTIONS.value,
            "25",
            com=lambda x: self.show_dr_hyperparameters(),
        )
        self.algo_dropdown.setChoices(UMLOptions.DR_ALGO_OPTIONS.value[0])
        self.show_dr_hyperparameters()

        settings_frm.grid(row=2, column=0, sticky=NW)
        self.scaling_dropdown.grid(row=0, column=0, sticky=NW)
        self.var_threshold_dropdown.grid(row=1, column=0, sticky=NW)
        self.algo_dropdown.grid(row=2, column=0, sticky=NW)
        self.main_frm.mainloop()

    def show_dr_hyperparameters(self):
        if hasattr(self, "hyperparameters_frm"):
            self.hyperparameters_frm.destroy()
            self.value_frm.destroy()
            self.run_frm.destroy()

        self.hyperparameters_frm = LabelFrame(
            self.main_frm,
            text="GRID SEARCH HYPER-PARAMETERS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.value_frm = LabelFrame(self.main_frm, fg="black")
        self.dr_value_entry_box = Entry_Box(self.value_frm, "VALUE: ", "12")
        self.value_frm.grid(row=3, column=0, sticky=NW)
        self.dr_value_entry_box.grid(row=0, column=1, sticky=NW)
        self.hyperparameters_frm.grid(row=4, column=0, sticky=NW)

        if self.algo_dropdown.getChoices() == Unsupervised.UMAP.value:
            Label(self.hyperparameters_frm, text=Unsupervised.N_NEIGHBORS.value).grid(
                row=1, column=0
            )
            Label(self.hyperparameters_frm, text=Unsupervised.MIN_DISTANCE.value).grid(
                row=1, column=1
            )
            Label(self.hyperparameters_frm, text=Unsupervised.SPREAD.value).grid(
                row=1, column=2
            )

            self.n_neighbors_estimators_listbox = Listbox(
                self.hyperparameters_frm, bg="lightgrey", fg="black", height=5, width=15
            )
            self.min_distance_listbox = Listbox(
                self.hyperparameters_frm, bg="lightgrey", fg="black", height=5, width=15
            )
            self.spread_listbox = Listbox(
                self.hyperparameters_frm, bg="lightgrey", fg="black", height=5, width=15
            )

            neighbours_add_btn = Button(
                self.hyperparameters_frm,
                text="ADD",
                fg="blue",
                command=lambda: self.add_to_listbox_from_entrybox(
                    list_box=self.n_neighbors_estimators_listbox,
                    entry_box=self.dr_value_entry_box,
                ),
            )
            min_distance_add_btn = Button(
                self.hyperparameters_frm,
                text="ADD",
                fg="blue",
                command=lambda: self.add_to_listbox_from_entrybox(
                    list_box=self.min_distance_listbox,
                    entry_box=self.dr_value_entry_box,
                ),
            )
            spread_add_btn = Button(
                self.hyperparameters_frm,
                text="ADD",
                fg="blue",
                command=lambda: self.add_to_listbox_from_entrybox(
                    list_box=self.spread_listbox, entry_box=self.dr_value_entry_box
                ),
            )

            neighbours_remove_btn = Button(
                self.hyperparameters_frm,
                text="REMOVE",
                fg="red",
                command=lambda: self.remove_from_listbox(
                    list_box=self.n_neighbors_estimators_listbox
                ),
            )
            min_distance_remove_btn = Button(
                self.hyperparameters_frm,
                text="REMOVE",
                fg="red",
                command=lambda: self.remove_from_listbox(
                    list_box=self.min_distance_listbox
                ),
            )
            spread_remove_btn = Button(
                self.hyperparameters_frm,
                text="REMOVE",
                fg="red",
                command=lambda: self.remove_from_listbox(list_box=self.spread_listbox),
            )

            self.add_values_to_several_listboxes(
                list_boxes=[
                    self.n_neighbors_estimators_listbox,
                    self.min_distance_listbox,
                    self.spread_listbox,
                ],
                values=[15, 0.1, 1],
            )

            neighbours_add_btn.grid(row=2, column=0)
            min_distance_add_btn.grid(row=2, column=1)
            spread_add_btn.grid(row=2, column=2)
            neighbours_remove_btn.grid(row=3, column=0)
            min_distance_remove_btn.grid(row=3, column=1)
            spread_remove_btn.grid(row=3, column=2)

            self.n_neighbors_estimators_listbox.grid(row=4, column=0, sticky=NW)
            self.min_distance_listbox.grid(row=4, column=1, sticky=NW)
            self.spread_listbox.grid(row=4, column=2, sticky=NW)
            self.create_run_frm(run_function=self.__run_umap_gridsearch)

        elif self.algo_dropdown.getChoices() == Unsupervised.TSNE.value:
            Label(self.hyperparameters_frm, text="PERPLEXITY").grid(row=1, column=0)
            self.perplexity_listbox = Listbox(
                self.hyperparameters_frm, bg="lightgrey", fg="black", height=5, width=15
            )
            perplexity_add_btn = Button(
                self.hyperparameters_frm,
                text="ADD",
                fg="blue",
                command=lambda: self.add_to_listbox_from_entrybox(
                    list_box=self.perplexity_listbox, entry_box=self.dr_value_entry_box
                ),
            )
            perplexity_remove_btn = Button(
                self.hyperparameters_frm,
                text="REMOVE",
                fg="red",
                command=lambda: self.remove_from_listbox(
                    list_box=self.perplexity_listbox
                ),
            )
            perplexity_add_btn.grid(row=2, column=0)
            perplexity_remove_btn.grid(row=3, column=0)
            self.perplexity_listbox.grid(row=4, column=0, sticky=NW)
            self.create_run_frm(run_function=self.__run_tsne_gridsearch)

    def __run_tsne_gridsearch(self):
        self.__get_settings()
        perplexities = [int(x) for x in self.perplexity_listbox.get(0, END)]
        if len(perplexities) == 0:
            raise NoSpecifiedOutputError("Provide value(s) for perplexity")
        hyperparameters = {
            "perplexity": perplexities,
            "scaler": self.scaling_dropdown.getChoices(),
            "variance": self.variance_selected,
        }
        tsne_searcher = TSNEGridSearch(
            data_path=self.data_path, save_dir=self.save_path
        )
        # tsne_searcher.fit(hyperparameters=hyperparameters)

    def __run_umap_gridsearch(self):
        self.__get_settings()
        n_neighbours = [
            float(x) for x in self.n_neighbors_estimators_listbox.get(0, END)
        ]
        min_distances = [float(x) for x in self.min_distance_listbox.get(0, END)]
        spreads = [float(x) for x in self.spread_listbox.get(0, END)]
        if len(min_distances) == 0 or len(n_neighbours) == 0 or len(spreads) == 0:
            raise NoSpecifiedOutputError(
                "Provide at least one hyperparameter value for neighbors, min distances, and spread"
            )
        hyper_parameters = {
            "n_neighbors": n_neighbours,
            "min_distance": min_distances,
            "spread": spreads,
            "scaler": self.scaling_dropdown.getChoices(),
            "variance": self.variance_selected,
        }
        umap_searcher = UmapEmbedder()
        threading.Thread(
            target=umap_searcher.fit(
                data_path=self.data_path,
                save_dir=self.save_path,
                hyper_parameters=hyper_parameters,
            )
        ).start()

    def __get_settings(self):
        self.variance_selected = (
            int(self.var_threshold_dropdown.getChoices()[:-1]) / 100
        )
        self.save_path = self.save_dir.folder_path
        self.data_path = self.dataset_file_selected.file_path
        self.scaler = self.scaling_dropdown.getChoices()
        check_if_dir_exists(self.save_path)
        check_file_exist_and_readable(self.data_path)


# _ = FitDimReductionPopUp(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/NG_Unsupervised/project_folder/project_config.ini')



__author__ = "Simon Nilsson"

import os.path
from typing import Dict, List, Optional, Union

try:
    from typing import Literal
except:
    from typing_extensions import Literal

import cv2

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_str, check_that_column_exist,
                                check_valid_boolean, check_valid_lst)
from simba.utils.enums import Dtypes
from simba.utils.errors import FrameRangeError
from simba.utils.printing import stdout_success
from simba.utils.read_write import (find_video_of_file, get_fn_ext,
                                    get_video_meta_data, read_df,
                                    read_frm_of_video)


class AnnotationFrameExtractor(ConfigReader):
    """
    Extracts all human annotated frames where behavior is annotated as present into images within a SimBA project.

    :param str config_path: path to SimBA configparser.ConfigParser project_config.in
    :param list clfs: Names of classifiers to extract behavior-present images from.
    :param List[Union[str, os.PathLike]] data_paths: List of files with annotations to extract images from.
    :param int downsample: How much to downsample each image resolution with. If None or 1, then no Downsample.
    :param bool greyscale: If True, saves the images as greyscale.

    :example:
    >>> extractor = AnnotationFrameExtractor(config_path='project_folder/project_config.ini', clfs=['Sniffing', 'Attack'], settings={'downsample': 2})
    >>> extractor.run()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 data_paths: List[Union[str, os.PathLike]],
                 clfs: List[str],
                 downsample: Optional[Union[float, int]] = None,
                 img_format: Optional[Literal['png', 'webp', 'jpg']] = 'png',
                 greyscale: Optional[bool] = False):

        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        check_valid_lst(data=clfs, valid_dtypes=(str,), valid_values=self.clf_names, min_len=1, source='classifiers')
        check_valid_lst(data=data_paths, valid_dtypes=(str,), min_len=1, source='data_paths')
        for i in data_paths: check_file_exist_and_readable(file_path=i)
        if downsample is not None:
            check_float(name='downsample', value=downsample, min_value=1)
        check_str(name='img_format', value=img_format, options=('png', 'webp', 'jpg'))
        check_valid_boolean(source='greyscale', value=greyscale)
        self.clfs = clfs
        self.downsample = downsample
        self.img_format = img_format
        self.greyscale = greyscale
        self.data_paths = data_paths

    def run(self):
        if not os.path.exists(self.annotated_frm_dir):
            os.makedirs(self.annotated_frm_dir)
        for file_cnt, file_path in enumerate(self.data_paths):
            _, video_name, _ = get_fn_ext(filepath=file_path)
            video_path = find_video_of_file(video_dir=self.video_dir, filename=video_name, raise_error=True)
            df = read_df(file_path=file_path, file_type=self.file_type)
            for clf in self.clfs:
                check_that_column_exist(df=df, column_name=clf, file_name=file_path)
            cap = cv2.VideoCapture(video_path)
            video_meta_data = get_video_meta_data(video_path=video_path)
            if isinstance(self.downsample, (int, float)) and self.downsample > 1:
                size = (int(video_meta_data['width'] / self.downsample), int(video_meta_data['height'] / self.downsample))
            else:
                size = None
            for clf in self.clfs:
                save_dir = os.path.join(self.annotated_frm_dir, video_name, clf)
                if not os.path.exists(save_dir): os.makedirs(save_dir)
                annot_idx = list(df.index[df[clf] == 1])
                for frm_cnt, frm in enumerate(annot_idx):
                    img = read_frm_of_video(video_path=cap, frame_index=frm, size=size, greyscale=self.greyscale)
                    cv2.imwrite(os.path.join(save_dir, f"{str(frm)}.{self.img_format}"), img)
                    print(f"Saved {clf} annotated img ({str(frm_cnt)}/{str(len(annot_idx))}), Video: {video_name} ...")

        self.timer.stop_timer()
        stdout_success(msg=f"Annotated frames saved in {self.annotated_frm_dir} directory", elapsed_time=self.timer.elapsed_time_str)


# test = AnnotationFrameExtractor(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                 clfs=['Attack'],
#                                 downsample=4,
#                                 data_paths=['/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/csv/targets_inserted/Together_1.csv'],
#                                 img_format='jpg',
#                                 greyscale=True)
# test.run()



__author__ = "Simon Nilsson"

import functools
import itertools
import multiprocessing
import os
import platform
from typing import Any, Dict, List, Optional, Tuple, Union

import cv2
import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.roi_tools.ROI_feature_analyzer import ROIFeatureCreator
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_keys_exist_in_dict, check_int,
                                check_valid_boolean, check_valid_lst,
                                check_video_and_data_frm_count_align)
from simba.utils.data import slice_roi_dict_for_video
from simba.utils.enums import Formats, TextOptions
from simba.utils.errors import (BodypartColumnNotFoundError, NoFilesFoundError,
                                ROICoordinatesNotFoundError)
from simba.utils.printing import stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder,
                                    find_core_cnt, get_fn_ext,
                                    get_video_meta_data, read_df,
                                    remove_a_folder)
from simba.utils.warnings import DuplicateNamesWarning

ROI_CENTERS = "roi_centers"
ROI_EAR_TAGS = "roi_ear_tags"
DIRECTIONALITY = "directionality"
DIRECTIONALITY_STYLE = "directionality_style"
BORDER_COLOR = "border_color"
POSE = "pose_estimation"
ANIMAL_NAMES = "animal_names"
STYLE_KEYS = [ROI_CENTERS,ROI_EAR_TAGS,DIRECTIONALITY,BORDER_COLOR,POSE,DIRECTIONALITY_STYLE,ANIMAL_NAMES,]


def _roi_feature_visualizer_mp(frm_range: Tuple[int, np.ndarray],
                               data_df: pd.DataFrame,
                               text_locations: dict,
                               font_size: float,
                               circle_size: Union[float, int],
                               save_temp_dir: str,
                               video_meta_data: dict,
                               shape_info: dict,
                               shape_names: list,
                               style_attr: dict,
                               video_path: str,
                               animal_names: list,
                               roi_dict: dict,
                               bp_lk: dict,
                               animal_bp_names: List[str],
                               animal_bp_dict: dict,
                               roi_features_df: pd.DataFrame,
                               directing_data: Union[pd.DataFrame, None]):
    def __insert_texts(shape_info: dict, img: np.ndarray):
        for shape_name, shape_info in shape_info.items():
            shape_color = shape_info["Color BGR"]
            for cnt, animal_data in bp_lk.items():
                animal, animal_bp, _ = animal_data
                animal_name = f"{animal} {animal_bp}"
                cv2.putText(img, text_locations[animal_name][shape_name]["in_zone_text"], text_locations[animal_name][shape_name]["in_zone_text_loc"], font, font_size, shape_color, 1)
                cv2.putText(img, text_locations[animal_name][shape_name]["distance_text"], text_locations[animal_name][shape_name]["distance_text_loc"], font, font_size, shape_color, 1)
                if directing_data is not None and style_attr[DIRECTIONALITY]:
                    cv2.putText(img, text_locations[animal][shape_name]["directing_text"], text_locations[animal][shape_name]["directing_text_loc"], font, font_size, shape_color, 1)
        return img

    fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
    font = cv2.FONT_HERSHEY_SIMPLEX
    group_cnt, frm_range = frm_range[0], frm_range[1]
    start_frm, current_frm, end_frm = frm_range[0], frm_range[0], frm_range[-1]
    save_path = os.path.join(save_temp_dir, f"{group_cnt}.mp4")
    writer = cv2.VideoWriter(save_path, fourcc, video_meta_data["fps"], (video_meta_data["width"] * 2, video_meta_data["height"]))
    cap = cv2.VideoCapture(video_path)
    cap.set(1, current_frm)
    while current_frm <= end_frm:
        ret, img = cap.read()
        if ret:
            img = cv2.copyMakeBorder(img, 0, 0, 0, int(video_meta_data["width"]), borderType=cv2.BORDER_CONSTANT, value=style_attr[BORDER_COLOR],)
            img = __insert_texts(shape_info=shape_info, img=img)
            if style_attr[POSE]:
                for animal_name, bp_data in animal_bp_dict.items():
                    for bp_cnt, bp in enumerate(zip(bp_data["X_bps"], bp_data["Y_bps"])):
                        bp_cords = data_df.loc[current_frm, list(bp)].values.astype(np.int64)
                        cv2.circle(img, (bp_cords[0], bp_cords[1]), 0, animal_bp_dict[animal_name]["colors"][bp_cnt], circle_size)
            if style_attr[ANIMAL_NAMES]:
                for animal_name, bp_data in animal_bp_dict.items():
                    headers = [bp_data["X_bps"][-1], bp_data["Y_bps"][-1]]
                    bp_cords = data_df.loc[current_frm, headers].values.astype(np.int64)
                    cv2.putText(img, animal_name, (bp_cords[0], bp_cords[1]), font, font_size, animal_bp_dict[animal_name]["colors"][0], 1)

            img = PlottingMixin.roi_dict_onto_img(img=img, roi_dict=roi_dict, circle_size=circle_size, show_tags=style_attr[ROI_EAR_TAGS], show_center=style_attr[ROI_CENTERS])

            for animal_name, shape_name in itertools.product(animal_bp_names, shape_names):
                in_zone_col_name = f"{shape_name} {animal_name} in zone"
                distance_col_name = f"{shape_name} {animal_name} distance"
                in_zone_value = str(bool(roi_features_df.loc[current_frm, in_zone_col_name]))
                distance_value = round(roi_features_df.loc[current_frm, distance_col_name], 2)
                cv2.putText(img, in_zone_value, text_locations[animal_name][shape_name]["in_zone_data_loc"], font, font_size, shape_info[shape_name]["Color BGR"], 1)
                cv2.putText(img, str(distance_value), text_locations[animal_name][shape_name]["distance_data_loc"], font, font_size, shape_info[shape_name]["Color BGR"], 1)
            if (directing_data is not None) and (style_attr[DIRECTIONALITY]):
                for animal_name, shape_name in itertools.product(animal_names, shape_names):
                    facing_col_name = f"{shape_name} {animal_name} facing"
                    facing_value = roi_features_df.loc[current_frm, facing_col_name]
                    cv2.putText(img, str(bool(facing_value)), text_locations[animal_name][shape_name]["directing_data_loc"], font, font_size, shape_info[shape_name]["Color BGR"], 1)
                    if facing_value:
                        img = PlottingMixin.insert_directing_line(directing_df=directing_data,
                                                                  img=img,
                                                                  shape_name=shape_name,
                                                                  animal_name=animal_name,
                                                                  frame_id=current_frm,
                                                                  color=shape_info[shape_name]["Color BGR"],
                                                                  thickness=shape_info[shape_name]["Thickness"],
                                                                  style=style_attr[DIRECTIONALITY_STYLE])
            writer.write(np.uint8(img))
            print(f"Multiprocessing frame: {current_frm} / {video_meta_data['frame_count']}  on core {group_cnt}...")
            current_frm += 1
        else:
            break
    writer.release()
    return group_cnt


class ROIfeatureVisualizerMultiprocess(ConfigReader):
    """
    Visualize features that depend on the relationships between the location of the animals and user-defined
    ROIs. E.g., distances to centroids of ROIs, cumulative time spent in ROIs, if animals are directing towards ROIs
    etc.

    :param Union[str, os.PathLike] config_path: Path to SimBA project config file in Configparser format
    :param Union[str, os.PathLike] video_path: Path to video file to overlay ROI features on.
    :param List[str] body_parts: List of body-parts to use as proxy for animal location(s).
    :param Dict[str, Any] style_attr: User-defined styles (sizes, colors etc.)
    :param Optional[int] core_cnt: Number of cores to use. Defaults to -1 which is all available cores.

    .. note:
       `Tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial.md#part-5-visualizing-roi-features>`__.

        See :meth:`simba.ROI_feature_visualizer.ROIfeatureVisualizer` for single process class. Would be slower but potentially more reliable.

    .. image:: _static/img/ROIfeatureVisualizer_1.png
       :width: 700
       :align: center

    .. image:: _static/img/ROIfeatureVisualizer_2.png
       :width: 700
       :align: center


    :example:
    >>> style_attr = {'roi_centers': True, 'roi_ear_tags': True, 'directionality': True, 'directionality_style': 'funnel', 'border_color': (0, 0, 0), 'pose_estimation': True, 'animal_names': True}
    >>> test = ROIfeatureVisualizerMultiprocess(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/spontenous_alternation/project_folder/project_config.ini',
    >>>                             video_path='/Users/simon/Desktop/envs/simba/troubleshooting/spontenous_alternation/project_folder/videos/NOR ENCODING FExMP8.mp4',
    >>>                             style_attr=style_attr,
    >>>                             body_parts=['Center'], core_cnt=-1)
    >>> test.run()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 video_path: Union[str, os.PathLike],
                 body_parts: List[str],
                 style_attr: Dict[str, Any],
                 core_cnt: int = -1,
                 gpu: bool = False):

        check_int(name=f"{self.__class__.__name__} core_cnt",value=core_cnt,min_value=-1,max_value=find_core_cnt()[0])
        if core_cnt == -1:
            core_cnt = find_core_cnt()[0]
        check_file_exist_and_readable(file_path=video_path)
        ConfigReader.__init__(self, config_path=config_path)
        PlottingMixin.__init__(self)
        check_if_keys_exist_in_dict(data=style_attr,key=STYLE_KEYS,name=f"{self.__class__.__name__} style_attr")
        check_valid_boolean(value=[gpu], source=f"{self.__class__.__name__} gpu", raise_error=True)
        self.gpu = gpu
        if not os.path.isfile(self.roi_coordinates_path):
            raise ROICoordinatesNotFoundError(expected_file_path=self.roi_coordinates_path)
        self.read_roi_data()
        _, self.video_name, _ = get_fn_ext(video_path)
        self.roi_dict, self.shape_names = slice_roi_dict_for_video(data=self.roi_dict, video_name=self.video_name)
        self.core_cnt, self.style_attr = core_cnt, style_attr
        self.save_path = os.path.join(self.roi_features_save_dir, f"{self.video_name}.mp4")
        if not os.path.exists(self.roi_features_save_dir):
            os.makedirs(self.roi_features_save_dir)
        self.save_temp_dir = os.path.join(self.roi_features_save_dir, "temp")
        if os.path.exists(self.save_temp_dir):
            remove_a_folder(folder_dir=self.save_temp_dir)
        os.makedirs(self.save_temp_dir)
        self.data_path = os.path.join(self.outlier_corrected_dir, f"{self.video_name}.{self.file_type}")
        if not os.path.isfile(self.data_path):
            raise NoFilesFoundError(msg=f"SIMBA ERROR: Could not find the file at path {self.data_path}. Make sure the data file exist to create ROI visualizations", source=self.__class__.__name__,)
        check_valid_lst(data=body_parts, source=f"{self.__class__.__name__} body-parts", valid_dtypes=(str,), min_len=1,)
        for bp in body_parts:
            if bp not in self.body_parts_lst:
                raise BodypartColumnNotFoundError(msg=f"The body-part {bp} is not a valid body-part in the SimBA project. Options: {self.body_parts_lst}",source=self.__class__.__name__,)
        self.roi_feature_creator = ROIFeatureCreator(config_path=config_path, body_parts=body_parts, append_data=False, data_path=self.data_path)
        self.roi_feature_creator.run()
        self.bp_lk = self.roi_feature_creator.bp_lk
        self.animal_names = [v[0] for v in self.bp_lk.values()]
        self.animal_bp_names = [f"{v[0]} {v[1]}" for v in self.bp_lk.values()]
        self.video_meta_data = get_video_meta_data(video_path, fps_as_int=False)
        self.fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        self.cap = cv2.VideoCapture(video_path)
        check_video_and_data_frm_count_align(video=video_path, data=self.data_path, name=video_path, raise_error=False)
        self.direct_viable = self.roi_feature_creator.roi_directing_viable
        self.data_df = read_df(file_path=self.data_path, file_type=self.file_type)
        self.shape_dicts = self.__create_shape_dicts()
        self.directing_df = self.roi_feature_creator.dr
        self.video_path = video_path
        self.roi_features_df = self.roi_feature_creator.out_df
        if platform.system() == "Darwin":
            multiprocessing.set_start_method("spawn", force=True)

    def __create_shape_dicts(self):
        shape_dicts = {}
        for shape, df in self.roi_dict.items():
            if not df["Name"].is_unique:
                df = df.drop_duplicates(subset=["Name"], keep="first")
                DuplicateNamesWarning(msg=f'Some of your ROIs with the same shape ({shape}) has the same names for video {self.video_name}. E.g., you have two rectangles named "My rectangle". SimBA prefers ROI shapes with unique names. SimBA will keep one of the unique shape names and drop the rest.', source=self.__class__.__name__,)
            d = df.set_index("Name").to_dict(orient="index")
            shape_dicts = {**shape_dicts, **d}
        return shape_dicts

    def __calc_text_locs(self):
        add_spacer = TextOptions.FIRST_LINE_SPACING.value
        self.loc_dict = {}
        txt_strs = []
        for animal_cnt, animal_bp_name in enumerate(self.animal_bp_names):
            for shape in self.shape_names:
                txt_strs.append(animal_bp_name + ' ' + shape + ' center distance')
        longest_text_str = str(max(txt_strs, key=len))
        self.font_size, x_spacer, y_spacer = PlottingMixin().get_optimal_font_scales(text=longest_text_str, accepted_px_width=int(self.video_meta_data["width"] / 2), accepted_px_height=int(self.video_meta_data["height"] / 15), text_thickness=3)
        self.circle_size = PlottingMixin().get_optimal_circle_size(frame_size=(int(self.video_meta_data["height"]), int(self.video_meta_data["height"])), circle_frame_ratio=100)
        for animal_cnt, animal_data in self.bp_lk.items():
            animal, animal_bp, _ = animal_data
            animal_name = f"{animal} {animal_bp}"
            self.loc_dict[animal_name] = {}
            self.loc_dict[animal] = {}
            for shape in self.shape_names:
                self.loc_dict[animal_name][shape] = {}
                self.loc_dict[animal_name][shape]["in_zone_text"] = f"{shape} {animal_name} in zone"
                self.loc_dict[animal_name][shape]["distance_text"] = f"{shape} {animal_name} distance"
                self.loc_dict[animal_name][shape]["in_zone_text_loc"] = ((self.video_meta_data["width"] + TextOptions.BORDER_BUFFER_X.value), (self.video_meta_data["height"] - (self.video_meta_data["height"] + 10) + y_spacer * add_spacer))
                self.loc_dict[animal_name][shape]["in_zone_data_loc"] = (int(self.video_meta_data["width"] + x_spacer + TextOptions.BORDER_BUFFER_X.value), (self.video_meta_data["height"] - (self.video_meta_data["height"] + 10) + y_spacer * add_spacer))
                add_spacer += 1
                self.loc_dict[animal_name][shape]["distance_text_loc"] = ((self.video_meta_data["width"] + TextOptions.BORDER_BUFFER_X.value), (self.video_meta_data["height"] - (self.video_meta_data["height"] + 10) + y_spacer * add_spacer))
                self.loc_dict[animal_name][shape]["distance_data_loc"] = (int(self.video_meta_data["width"] + x_spacer + TextOptions.BORDER_BUFFER_X.value), (self.video_meta_data["height"] - (self.video_meta_data["height"] + 10) + y_spacer * add_spacer))
                add_spacer += 1
                if self.direct_viable and self.style_attr[DIRECTIONALITY]:
                    self.loc_dict[animal][shape] = {}
                    self.loc_dict[animal][shape]["directing_text"] = f"{shape} {animal} facing"
                    self.loc_dict[animal][shape]["directing_text_loc"] = ((self.video_meta_data["width"] + TextOptions.BORDER_BUFFER_X.value), (self.video_meta_data["height"] - (self.video_meta_data["height"] + 10) + y_spacer * add_spacer))
                    self.loc_dict[animal][shape]["directing_data_loc"] = (int(self.video_meta_data["width"] + x_spacer + TextOptions.BORDER_BUFFER_X.value), (self.video_meta_data["height"] - (self.video_meta_data["height"] + 10) + y_spacer * add_spacer))
                    add_spacer += 1

    def __get_border_img_size(self, video_path: Union[str, os.PathLike]):
        cap = cv2.VideoCapture(video_path)
        cap.set(1, 1)
        _, img = self.cap.read()
        bordered_img = cv2.copyMakeBorder(img, 0, 0, 0, int(self.video_meta_data["width"]), borderType=cv2.BORDER_CONSTANT, value=[0, 0, 0])
        cap.release()
        return bordered_img.shape[0], bordered_img.shape[1]

    def run(self):
        self.img_w_border_h, self.img_w_border_w = self.__get_border_img_size(video_path=self.video_path)
        self.__calc_text_locs()
        frm_lst = np.arange(0, len(self.data_df) + 1)
        frm_lst = np.array_split(frm_lst, self.core_cnt)
        frame_range = []
        for i in range(len(frm_lst)): frame_range.append((i, frm_lst[i]))
        print(f"Creating ROI feature images, multiprocessing (chunksize: {self.multiprocess_chunksize}, cores: {self.core_cnt})...")
        with multiprocessing.Pool(self.core_cnt, maxtasksperchild=self.maxtasksperchild) as pool:
            constants = functools.partial(_roi_feature_visualizer_mp,
                                          data_df=self.data_df.reset_index(drop=True),
                                          text_locations=self.loc_dict,
                                          font_size=self.font_size,
                                          circle_size=self.circle_size,
                                          video_meta_data=self.video_meta_data,
                                          shape_info=self.shape_dicts,
                                          roi_dict=self.roi_dict,
                                          style_attr=self.style_attr,
                                          save_temp_dir=self.save_temp_dir,
                                          directing_data=self.directing_df,
                                          shape_names=self.shape_names,
                                          animal_bp_names=self.animal_bp_names,
                                          video_path=self.video_path,
                                          animal_names=self.animal_names,
                                          animal_bp_dict=self.animal_bp_dict,
                                          bp_lk=self.bp_lk,
                                          roi_features_df=self.roi_features_df.reset_index(drop=True))

            for cnt, result in enumerate(pool.imap(constants, frame_range, chunksize=self.multiprocess_chunksize)):
               print(f"Batch core {result+1}/{self.core_cnt} complete...")
            print(f"Joining {self.video_name} multi-processed video...")
            concatenate_videos_in_folder(in_folder=self.save_temp_dir, save_path=self.save_path, video_format="mp4", remove_splits=True, gpu=self.gpu)
            self.timer.stop_timer()
            pool.terminate()
            pool.join()
            stdout_success(msg=f"Video {self.video_name} complete. Video saved in directory {self.roi_features_save_dir}.", elapsed_time=self.timer.elapsed_time_str)



# if __name__ == '__main__':
#     style_attr = {'roi_centers': True, 'roi_ear_tags': True, 'directionality': False, 'directionality_style': 'funnel', 'border_color': (0, 0, 0), 'pose_estimation': True, 'animal_names': False}
#     test = ROIfeatureVisualizerMultiprocess(config_path=r"C:\troubleshooting\roi_feature_issue\project_folder\project_config.ini",
#                                                 video_path=r"C:\troubleshooting\roi_feature_issue\project_folder\videos\20250130_Oxyipn_Vls_D4_Sst-107.avi",
#                                                 style_attr=style_attr,
#                                                 body_parts=['nose'],
#                                                 core_cnt=-1)
#     test.run()




# if __name__ == '__main__':
#     style_attr = {'roi_centers': True, 'roi_ear_tags': True, 'directionality': True, 'directionality_style': 'funnel', 'border_color': (0, 0, 0), 'pose_estimation': True, 'animal_names': True}
#     test = ROIfeatureVisualizerMultiprocess(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini",
#                                                 video_path=r"C:\troubleshooting\mitra\project_folder\videos\503_MA109_Gi_CNO_0514.mp4",
#                                                 style_attr=style_attr,
#                                                 body_parts=['Center'],
#                                                 core_cnt=-1)
#     test.run()
#






# if __name__ == '__main__':
#     style_attr = {'roi_centers': True, 'roi_ear_tags': True, 'directionality': True, 'directionality_style': 'funnel', 'border_color': (0, 0, 0), 'pose_estimation': True, 'animal_names': False}
#     test = ROIfeatureVisualizerMultiprocess(config_path=r"C:\troubleshooting\platea\project_folder\project_config.ini",
#                               video_path=r"C:\troubleshooting\platea\project_folder\videos\Video_1.mp4",
#                               body_parts=['NOSE'],
#                               style_attr=style_attr)
#     test.run()



# style_attr = {'roi_centers': True, 'roi_ear_tags': True, 'directionality': True, 'directionality_style': 'funnel', 'border_color': (0, 0, 0), 'pose_estimation': True, 'animal_names': True}
# test = ROIfeatureVisualizerMultiprocess(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/spontenous_alternation/project_folder/project_config.ini',
#                             video_path='/Users/simon/Desktop/envs/simba/troubleshooting/spontenous_alternation/project_folder/videos/NOR ENCODING FExMP8.mp4',
#                             style_attr=style_attr,
#                             body_parts=['Center'], core_cnt=-1)
# test.run()


# style_attr = {'roi_centers': True, 'roi_ear_tags': True, 'directionality': True, 'directionality_style': 'funnel', 'border_color': (0, 0, 0), 'pose_estimation': True, 'animal_names': True}
# test = ROIfeatureVisualizerMultiprocess(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/RAT_NOR/project_folder/project_config.ini',
#                             video_path='/Users/simon/Desktop/envs/simba/troubleshooting/RAT_NOR/project_folder/videos/2022-06-20_NOB_DOT_4.mp4',
#                             style_attr=style_attr,
#                             body_parts=['Nose'], core_cnt=-1)
# test.run()
#

# style_attr = {'ROI_centers': True,
#               'ROI_ear_tags': True,
#               'Directionality': True,
#               'Directionality_style': 'Funnel',
#               'Border_color': (0, 128, 0),
#               'Pose_estimation': True,
#               'Directionality_roi_subset': ['My_polygon']}

# roi_feature_visualizer = ROIfeatureVisualizerMultiprocess(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                                           video_name='Together_1.avi',
#                                                           style_attr=style_attr,
#                                                           core_cnt=3)
# roi_feature_visualizer.run()


# style_attr = {'ROI_centers': True,
#               'ROI_ear_tags': True,
#               'Directionality': True,
#               'Directionality_style': 'Funnel',
#               'Border_color': (0, 128, 0),
#               'Pose_estimation': True}
# roi_feature_visualizer = ROIfeatureVisualizerMultiprocess(config_path='/Users/simon/Desktop/envs/simba_dev/tests/test_data/mouse_open_field/project_folder/project_config.ini',
#                                                           video_name='Video1.mp4',
#                                                           style_attr=style_attr,
#                                                           core_cnt=3)
# roi_feature_visualizer.create_visualization()
#
# style_attr = {'ROI_centers': True, 'ROI_ear_tags': True, 'Directionality': True, 'Directionality_style': 'Funnel', 'Border_color': (0, 128, 0), 'Pose_estimation': True}
# test = ROIfeatureVisualizerMultiprocess(config_path='/Users/simon/Desktop/envs/simba_dev/tests/test_data/mouse_open_field/project_folder/project_config.ini', video_name='Video1.mp4', style_attr=style_attr, core_cnt=5)
# test.create_visualization()



import json
import os
from datetime import datetime
from typing import Optional, Tuple, Union

import pandas as pd

try:
    from typing import Literal
except:
    from typing_extensions import Literal

import numpy as np

from simba.mixins.image_mixin import ImageMixin
from simba.third_party_label_appenders.transform.utils import (
    arr_to_b64, b64_to_arr, normalize_img_dict, scale_pose_img_sizes)
from simba.utils.checks import (check_if_dir_exists,
                                check_if_keys_exist_in_dict, check_str,
                                check_valid_boolean, check_valid_tuple)
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    img_array_to_clahe)


class LabelMe2DataFrame:
    """
    Convert a directory of labelme .json files into a pandas dataframe.

    .. note::
       The images are stores as a 64-bit bytestring under the ``image`` header of the output dataframe.

    :param Union[str, os.PathLike] labelme_dir: Directory with labelme json files.
    :param Optional[bool] greyscale: If True, converts the labelme images to greyscale if in rgb format. Default: False.
    :param Optional[bool] pad: If True, checks if all images are the same size and if not; pads the images with black border so all images are the same size.
    :param Union[Literal['min', 'max'], Tuple[int, int]] size: The size of the output images. Can be the smallesgt (min) the largest (max) or a tuple with the width and height of the images. Automatically corrects the labels to account for the image size.
    :param Optional[bool] normalize: If true, normalizes the images. Default: False.
    :param Optional[Union[str, os.PathLike]] save_path: The location where to store the dataframe. If None, then returns the dataframe. Default: None.

    :rtype: Union[None, pd.DataFrame]

    :example I:
    >>> LABELME_DIR = r'C:\troubleshooting\coco_data\labels\test_2'
    >>> runner = LabelMe2DataFrame(labelme_dir=LABELME_DIR)
    >>> runner.run()

    :example II:
    >>> LABELME_DIR = r'C:\troubleshooting\coco_data\labels\test_2'
    >>> runner = LabelMe2DataFrame(labelme_dir=LABELME_DIR, greyscale=True, pad=True, normalize=True, size='min')
    >>> runner.run()

    """

    def __init__(self,
                 labelme_dir: Union[str, os.PathLike],
                 greyscale: Optional[bool] = False,
                 clahe: bool = False,
                 pad: Optional[bool] = False,
                 size: Union[Literal['min', 'max'], Tuple[int, int]] = None,
                 normalize: bool = False,
                 save_path: Optional[Union[str, os.PathLike]] = None,
                 verbose: bool = True):

        if save_path is not None:
            if os.path.isdir(save_path):
                save_path = os.path.join(save_path, f'labelme_data_{datetime.now().strftime("%Y%m%d%H%M%S")}.csv')
        save_path = save_path
        if size is not None:
            if isinstance(size, tuple):
                check_valid_tuple(source=f'{self.__class__.__name__} size', accepted_lengths=(2,), valid_dtypes=(int,), x=size)
            elif isinstance(size, str):
                check_str(name=f'{self.__class__.__name__} size', value=size, options=('min', 'max',), raise_error=True)
        check_if_dir_exists(in_dir=labelme_dir)
        check_valid_boolean(value=greyscale, source=f'{self.__class__.__name__} greyscale', raise_error=True)
        check_valid_boolean(value=pad, source=f'{self.__class__.__name__} pad', raise_error=True)
        check_valid_boolean(value=normalize, source=f'{self.__class__.__name__} normalize', raise_error=True)
        check_valid_boolean(value=clahe, source=f'{self.__class__.__name__} clahe', raise_error=True)
        check_valid_boolean(value=verbose, source=f'{self.__class__.__name__} verbose', raise_error=True)
        self.annotation_paths = find_files_of_filetypes_in_directory(directory=labelme_dir, extensions=['.json'], raise_error=True)
        self.greyscale, self.pad, self.size = greyscale, pad, size
        self.normalize, self.verbose, self.clahe = normalize, verbose, clahe
        self.labelme_dir, self.save_path = labelme_dir, save_path


    def run(self):
        timer = SimbaTimer(start=True)
        images, annotations = {}, []
        for cnt, annot_path in enumerate(self.annotation_paths):
            if self.verbose:
                print(f'Reading annotation file {cnt + 1}/{len(self.annotation_paths)}...')
            with open(annot_path) as f:
                annot_data = json.load(f)
            check_if_keys_exist_in_dict(data=annot_data, key=['shapes', 'imageData'], name=annot_path)
            img_name = annot_data['imagePath']
            images[img_name] = b64_to_arr(annot_data['imageData'])
            if self.greyscale:
                images[img_name] = ImageMixin.img_to_greyscale(img=images[img_name])
            if self.clahe:
                images[img_name] = img_array_to_clahe(img=images[img_name])
            img_data = {}
            for bp_data in annot_data['shapes']:
                check_if_keys_exist_in_dict(data=bp_data, key=['label', 'points'], name=annot_path)
                point_x, point_y = bp_data['points'][0], bp_data['points'][1]
                lbl = bp_data['label']
                img_data[f'{lbl}_x'], img_data[f'{lbl}_y'] = point_x, point_y
            img_data['image_name'] = img_name
            annotations.append(pd.DataFrame.from_dict(img_data, orient='index').T)
        if self.pad:
            if self.verbose: print('Padding images...')
            images = ImageMixin.pad_img_stack(image_dict=images)
        if self.normalize:
            if self.verbose: print('Normalizing images...')
            images = normalize_img_dict(img_dict=images)
        img_lst = []
        for k, v in images.items():
            img_lst.append(arr_to_b64(v))
        out = pd.concat(annotations).reset_index(drop=True)
        out['image'] = img_lst
        if self.size is not None:
            if self.verbose: print('Resizing images...')
            pose_data = out.drop(['image', 'image_name'], axis=1)
            pose_data_arr = pose_data.values.reshape(-1, int(pose_data.shape[1] / 2), 2).astype(np.float32)
            new_pose, out['image'] = scale_pose_img_sizes(pose_data=pose_data_arr, imgs=list(out['image']), size=self.size)
            new_pose = new_pose.reshape(pose_data.shape[0], pose_data.shape[1])
            out.iloc[:, : new_pose.shape[1]] = new_pose
        if self.save_path is None:
            return out
        else:
            if self.verbose: print('Saving CSV file...')
            out.to_csv(self.save_path)
            timer.stop_timer()
            if self.verbose:
                stdout_success(msg=f'Labelme CSV file saved at {self.save_path}', elapsed_time=timer.elapsed_time_str, source=self.__class__.__name__)


# LABELME_DIR = r'C:\troubleshooting\coco_data\labels\test_2'
# runner = LabelMe2DataFrame(labelme_dir=LABELME_DIR)
#
#
# #runner.run()
#
# LABELME_DIR = r'C:\troubleshooting\coco_data\labels\test_2'
# runner = LabelMe2DataFrame(labelme_dir=LABELME_DIR, greyscale=True, pad=True, normalize=True, size='max', save_path=r'D:\imgs')
# runner.run()
#labelme_to_df(labelme_dir=r'C:\troubleshooting\coco_data\labels\test_2').run()
#    >>> labelme_to_df(labelme_dir=r'C:\troubleshooting\coco_data\labels\test_read', greyscale=False, pad=False, normalize=False, size='min').run()


import itertools
import os
from copy import deepcopy
from typing import Any, Dict, Optional, Union

import numpy as np
import pandas as pd
from sklearn.ensemble import RandomForestClassifier
from sklearn.inspection import permutation_importance

from simba.mixins.config_reader import ConfigReader
from simba.mixins.train_model_mixin import TrainModelMixin
from simba.mixins.unsupervised_mixin import UnsupervisedMixin
from simba.plotting.shap_agg_stats_visualizer import \
    ShapAggregateStatisticsCalculator
from simba.unsupervised.enums import Clustering, Unsupervised
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_keys_exist_in_dict, check_instance)
from simba.utils.enums import Formats
from simba.utils.errors import InvalidInputError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_unique_values_in_iterable, read_pickle

FEATURE_NAME = "FEATURE NAME"
FEATURE_IMPORTANCE = "IMPORTANCE"
F_STATISTIC = "F-STATISTIC"
MEASURE = "MEASURE"
P_VALUE = "P-VALUE"
CLUSTER = "CLUSTER"
PAIRED = "Paired clusters"
ONE_AGAINST_ALL = "One-against-all"
CORRELATION_METHODS = "correlation_methods"
GINI_IMPORTANCE = "gini_importance"
TUKEY = "tukey_posthoc"
METHOD = "method"
TARGET = "TARGET"
PEARSON = "pearson"
KENDALL = "kendall"
SHAP = "shap"
PLOTS = "plots"
CREATE = "create"
RUN = "run"
SAMPLE = "sample"
SPEARMAN = "spearman"
MEAN = "MEAN"
STDEV = "STANDARD DEVIATION"
PERMUTATION_IMPORTANCE = "permutation_importance"
DESCRIPTIVE_STATISTICS = "descriptive_statistics"
ANOVA_HEADERS = ["FEATURE NAME", "F-STATISTIC", "P-VALUE"]


class ClusterXAICalculator(UnsupervisedMixin, ConfigReader):
    """
    Class for building RF models on top of cluster assignments, and calculating latent space explainability metrics based on RF models.

    :param Union[str, os.PathLike] config_path: path to SimBA configparser.ConfigParser project_config.ini
    :param Union[str, os.PathLike] data_path: path to pickle holding unsupervised results in ``data_map.yaml`` format.
    :param Dict[str, Any] settings: Dict holding which explainability tests to use.

    :Example:
    >>> settings = {'gini_importance': True, 'permutation_importance': True, 'shap': {'method': 'cluster_paired', 'create': True, 'sample': 100}}
    >>> calculator = ClusterXAICalculator(config_path='unsupervised/project_folder/project_config.ini', data_path='unsupervised/cluster_models/quizzical_rhodes.pickle', settings=settings)
    >>> calculator.run()
    """

    def __init__(
        self,
        data_path: Union[str, os.PathLike],
        config_path: Union[str, os.PathLike],
        settings: Dict[str, Any],
    ):
        check_file_exist_and_readable(file_path=data_path)
        check_file_exist_and_readable(file_path=config_path)
        check_instance(
            source=f"{self.__class__.__name__} settings",
            instance=settings,
            accepted_types=(dict,),
        )
        ConfigReader.__init__(self, config_path=config_path)
        UnsupervisedMixin.__init__(self)
        self.settings, self.data_path = settings, data_path
        self.data = read_pickle(data_path=self.data_path)
        check_if_keys_exist_in_dict(
            data=self.data,
            key=[Unsupervised.METHODS.value, Clustering.CLUSTER_MODEL.value],
            name=self.data_path,
        )
        check_if_keys_exist_in_dict(
            data=self.settings,
            key=[GINI_IMPORTANCE, PERMUTATION_IMPORTANCE, SHAP],
            name=self.data_path,
        )
        self.save_path = os.path.join(
            self.logs_path,
            f"cluster_xai_statistics_{self.data[Unsupervised.DR_MODEL.value][Unsupervised.HASHED_NAME.value]}_{self.datetime}.{Formats.XLXS.value}",
        )

    def run(self):
        self.x_df = self.data[Unsupervised.METHODS.value][
            Unsupervised.SCALED_DATA.value
        ]
        self.cluster_data = self.data[Clustering.CLUSTER_MODEL.value][
            Unsupervised.MODEL.value
        ].labels_
        self.x_y_df = pd.concat(
            [
                self.x_df,
                pd.DataFrame(
                    self.cluster_data, columns=[CLUSTER], index=self.x_df.index
                ),
            ],
            axis=1,
        )
        self.mdl_file_name = self.data[Clustering.CLUSTER_MODEL.value][
            Unsupervised.HASHED_NAME.value
        ]
        self.cluster_cnt = get_unique_values_in_iterable(
            data=self.cluster_data,
            name=self.mdl_file_name,
            min=2,
        )

        with pd.ExcelWriter(self.save_path, mode="w") as writer:
            pd.DataFrame().to_excel(writer, sheet_name=" ", index=True)

        if (
            self.settings[GINI_IMPORTANCE]
            or self.settings[PERMUTATION_IMPORTANCE]
            or (self.settings[SHAP][METHOD] == PAIRED)
        ):
            self.__train_paired_rf_models()
        if self.settings[GINI_IMPORTANCE]:
            self.__gini_importance()
        if self.settings[PERMUTATION_IMPORTANCE]:
            self.__permutation_importance()
        if self.settings[SHAP][RUN]:
            self.__shap_values()
        self.timer.stop_timer()
        stdout_success(
            msg=f"Cluster XAI complete. Data saved at {self.save_path}",
            elapsed_time=self.timer.elapsed_time_str,
        )

    def __save_results(self, df: pd.DataFrame, name: str):
        with pd.ExcelWriter(self.save_path, mode="a") as writer:
            df.to_excel(writer, sheet_name=name, index=True)

    def __train_paired_rf_models(self, n_estimators: Optional[int] = 100):
        print("Training RF ML model...")
        clf = RandomForestClassifier(
            n_estimators=n_estimators,
            max_features="sqrt",
            n_jobs=-1,
            criterion="gini",
            min_samples_leaf=1,
            bootstrap=True,
            verbose=1,
        )
        self.rf_data = {}
        cluster_permutations = list(
            itertools.permutations(list(self.x_y_df[CLUSTER].unique()), 2)
        )
        for clf_cnt, (x1, x2) in enumerate(cluster_permutations):
            print(f"Training model {clf_cnt + 1}/{len(cluster_permutations)} ...")
            self.rf_data[x1] = {}
            target_df = self.x_y_df[self.x_y_df[CLUSTER] == x1].drop([CLUSTER], axis=1)
            non_target_df = self.x_y_df[self.x_y_df[CLUSTER] == x2].drop(
                [CLUSTER], axis=1
            )
            target_df[TARGET] = 1
            non_target_df[TARGET] = 0
            self.rf_data[x1]["X"] = pd.concat(
                [target_df, non_target_df], axis=0
            ).reset_index(drop=True)
            self.rf_data[x1]["Y"] = self.rf_data[x1]["X"].pop(TARGET)
            clf = TrainModelMixin().clf_fit(
                clf=clf, x_df=self.rf_data[x1]["X"], y_df=self.rf_data[x1]["Y"]
            )
            self.rf_data[x1][Unsupervised.MODEL.value] = deepcopy(clf)

    def __gini_importance(self):
        print("Calculating cluster gini importances'...")
        timer = SimbaTimer(start=True)
        for cluster_id, cluster_data in self.rf_data.items():
            importances = list(
                cluster_data[Unsupervised.MODEL.value].feature_importances_
            )
            gini_data = [
                (feature, round(importance, 6))
                for feature, importance in zip(
                    self.data[Unsupervised.METHODS.value][
                        Unsupervised.FEATURE_NAMES.value
                    ],
                    importances,
                )
            ]
            df = (
                pd.DataFrame(gini_data, columns=[FEATURE_NAME, FEATURE_IMPORTANCE])
                .sort_values(by=[FEATURE_IMPORTANCE], ascending=False)
                .reset_index(drop=True)
            )
            self.__save_results(df=df, name=f"GINI CLUSTER {str(cluster_id)}")
        timer.stop_timer()
        stdout_success(
            msg=f"Cluster features gini importances' complete",
            elapsed_time=timer.elapsed_time_str,
        )

    def __permutation_importance(self):
        print("Calculating permutation importance...")
        timer = SimbaTimer(start=True)
        for cluster_id, cluster_data in self.rf_data.items():
            p_importances = permutation_importance(
                cluster_data[Unsupervised.MODEL.value],
                cluster_data["X"],
                cluster_data["Y"],
                n_repeats=5,
                random_state=0,
            )
            df = pd.DataFrame(
                np.column_stack(
                    [
                        self.data[Unsupervised.METHODS.value][
                            Unsupervised.FEATURE_NAMES.value
                        ],
                        p_importances.importances_mean,
                        p_importances.importances_std,
                    ]
                ),
                columns=[FEATURE_NAME, MEAN, STDEV],
            )
            df = df.sort_values(by=[MEAN], ascending=False).reset_index(drop=True)
            self.__save_results(df=df, name=f"PERMUTATION CLUSTER {str(cluster_id)}")
        timer.stop_timer()
        stdout_success(
            msg=f"Cluster features permutation importances' complete",
            elapsed_time=timer.elapsed_time_str,
        )

    def __train_all_against_one_rf_models(self, n_estimators: Optional[int] = 100):
        """Private helper to create random forest classifiers training observations in one cluster against all observations in other clusters"""
        all_against_one_rf_mdls = {}
        for cluster_id in sorted(self.x_y_df[CLUSTER].unique()):
            clf = RandomForestClassifier(
                n_estimators=n_estimators,
                max_features="sqrt",
                n_jobs=-1,
                criterion="gini",
                min_samples_leaf=1,
                bootstrap=True,
                verbose=1,
            )
            cluster_df = self.x_y_df[self.x_y_df[CLUSTER] == cluster_id].drop(
                [CLUSTER], axis=1
            )
            noncluster_df = self.x_y_df[self.x_y_df[CLUSTER] != cluster_id].drop(
                [CLUSTER], axis=1
            )
            cluster_df[TARGET] = 1
            noncluster_df[TARGET] = 0
            x = pd.concat([cluster_df, noncluster_df], axis=0)
            y = x.pop(TARGET)
            all_against_one_rf_mdls[cluster_id] = TrainModelMixin().clf_fit(
                clf=clf, x_df=x, y_df=y
            )
        return all_against_one_rf_mdls

    def __shap_values(self):
        if self.settings[SHAP][METHOD] == PAIRED:
            print("Computing paired-clusters shap values ...")
            timer = SimbaTimer(start=True)
            cluster_combinations = list(
                itertools.combinations(list(self.rf_data.keys()), 2)
            )
            for cluster_one_id, cluster_two_id in cluster_combinations:
                mdl_name = f"SHAP CLUSTER {str(cluster_one_id)} vs. {str(cluster_two_id)} {self.mdl_file_name}"
                print(f"Computing {mdl_name} values ...")
                cluster_one_sample = self.x_y_df[
                    self.x_y_df[CLUSTER] == cluster_one_id
                ].drop(CLUSTER, axis=1)
                cluster_two_sample = self.x_y_df[
                    self.x_y_df[CLUSTER] == cluster_two_id
                ].drop(CLUSTER, axis=1)
                sample_n = min(
                    self.settings[SHAP][SAMPLE],
                    len(cluster_one_sample),
                    len(cluster_two_sample),
                )
                cluster_one_sample = cluster_one_sample.sample(sample_n, replace=False)
                cluster_two_sample = cluster_two_sample.sample(sample_n, replace=False)
                cluster_one_sample[mdl_name] = 1
                cluster_two_sample[mdl_name] = 0
                x_df = pd.concat(
                    [cluster_one_sample, cluster_two_sample], axis=0
                ).reset_index(drop=True)
                y_df = x_df.pop(mdl_name)
                shap_df, _, expected_value = TrainModelMixin().create_shap_log_mp(
                    ini_file_path=self.config_path,
                    rf_clf=self.rf_data[cluster_one_id]["MODEL"],
                    x_df=x_df,
                    y_df=y_df,
                    x_names=list(x_df.columns),
                    clf_name=mdl_name,
                    cnt_present=sample_n,
                    cnt_absent=sample_n,
                )
                _ = ShapAggregateStatisticsCalculator(
                    config_path=self.config_path,
                    classifier_name=mdl_name,
                    shap_df=shap_df,
                    shap_baseline_value=expected_value,
                    save_path=None,
                )
                cluster_one_shap = shap_df[shap_df[mdl_name] == 1]
                cluster_two_shap = shap_df[shap_df[mdl_name] == 0]
                mean_df_cluster_one, stdev_df_cluster_one = pd.DataFrame(
                    cluster_one_shap.mean(), columns=["MEAN"]
                ), pd.DataFrame(cluster_one_shap.std(), columns=["STDEV"])
                mean_df_cluster_two, stdev_df_cluster_two = pd.DataFrame(
                    cluster_two_shap.mean(), columns=["MEAN"]
                ), pd.DataFrame(cluster_two_shap.std(), columns=["STDEV"])
                mean_df_cluster_two["MEAN"] = mean_df_cluster_two["MEAN"] * -1
                results_cluster_two = (
                    mean_df_cluster_two.join(stdev_df_cluster_two)
                    .sort_values(by="MEAN", ascending=False)
                    .drop(
                        [mdl_name, "Expected_value", "Sum", "Prediction_probability"],
                        axis=0,
                    )
                )
                results_cluster_one = (
                    mean_df_cluster_one.join(stdev_df_cluster_one)
                    .sort_values(by="MEAN", ascending=False)
                    .drop(
                        [mdl_name, "Expected_value", "Sum", "Prediction_probability"],
                        axis=0,
                    )
                )
                self.__save_results(df=results_cluster_one, name=mdl_name)
                self.__save_results(df=results_cluster_two, name=mdl_name)
            timer.stop_timer()
            stdout_success(
                msg=f"Paired clusters SHAP values complete",
                elapsed_time=timer.elapsed_time_str,
            )

        elif self.settings[SHAP][METHOD] == ONE_AGAINST_ALL:
            timer = SimbaTimer(start=True)
            print("Calculating one-against-all shap values ...")
            mdls = self.__train_all_against_one_rf_models()
            for cluster_id, cluster_mdl in mdls.items():
                shap_mdl_name = (
                    f"SHAP CLUSTER {cluster_id} vs. ALL {self.mdl_file_name}"
                )
                print(f"Computing SHAP for cluster {shap_mdl_name}...")
                cluster_one_sample = self.x_y_df[
                    self.x_y_df[CLUSTER] == cluster_id
                ].drop(CLUSTER, axis=1)
                cluster_two_sample = self.x_y_df[
                    self.x_y_df[CLUSTER] != cluster_id
                ].drop(CLUSTER, axis=1)
                sample_n = min(
                    self.settings[SHAP][SAMPLE],
                    len(cluster_one_sample),
                    len(cluster_one_sample),
                )
                cluster_one_sample = cluster_one_sample.sample(sample_n, replace=False)
                cluster_two_sample = cluster_two_sample.sample(sample_n, replace=False)
                cluster_one_sample[shap_mdl_name] = 1
                cluster_two_sample[shap_mdl_name] = 0
                x_df = pd.concat(
                    [cluster_one_sample, cluster_two_sample], axis=0
                ).reset_index(drop=True)
                y_df = x_df.pop(shap_mdl_name)
                shap_df, _, expected_value = TrainModelMixin().create_shap_log_mp(
                    ini_file_path=self.config_path,
                    rf_clf=mdls[cluster_id],
                    x_df=x_df,
                    y_df=y_df,
                    x_names=list(x_df.columns),
                    clf_name=shap_mdl_name,
                    cnt_present=sample_n,
                    cnt_absent=sample_n,
                )
                _ = ShapAggregateStatisticsCalculator(
                    config_path=self.config_path,
                    classifier_name=shap_mdl_name,
                    shap_df=shap_df,
                    shap_baseline_value=expected_value,
                    save_path=None,
                )
                mean_df, stdev_df = pd.DataFrame(
                    shap_df.mean(), columns=["MEAN"]
                ), pd.DataFrame(shap_df.std(), columns=["STDEV"])
                shap_df = (
                    mean_df.join(stdev_df)
                    .sort_values(by="MEAN", ascending=False)
                    .drop(
                        [
                            shap_mdl_name,
                            "Expected_value",
                            "Sum",
                            "Prediction_probability",
                        ],
                        axis=0,
                    )
                )
                self.__save_results(df=shap_df, name=shap_mdl_name)
                timer.stop_timer()
                stdout_success(
                    msg=f"SHAP one-vs-all values complete",
                    elapsed_time=timer.elapsed_time_str,
                )

        else:
            raise InvalidInputError(
                msg=f"Shap parameter {self.settings[SHAP][METHOD]} not recognized",
                source=self.__class__.__name__,
            )


# settings = {"gini_importance": False, "permutation_importance": False, "shap": {"method": PAIRED, "run": True, "sample": 10}}
# calculator = ClusterXAICalculator(
#     config_path="/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/project_config.ini",
#     data_path="/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/error_xai/determined_wing.pickle",
#     settings=settings)
# # #
# # #
# calculator.run()



__author__ = "Simon Nilsson"

import os
from pathlib import Path
from tkinter import *
from typing import Union

from simba.data_processors.interpolate import Interpolate
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import (DropDownMenu, FileSelect, FolderSelect,
                                        SimbaButton)
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists)
from simba.utils.enums import Formats, Options
from simba.utils.read_write import str_2_bool

INTERPOLATOR_METHOD = {'MISSING BODY-PARTS': 'body-parts', 'MISSING ANIMALS': 'animals'}

class InterpolatePopUp(PopUpMixin, ConfigReader):
    """
    :example:
    >>> InterpolatePopUp(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')

    """

    def __init__(self, config_path: Union[str, os.PathLike]):
        PopUpMixin.__init__(self, title="INTERPOLATE POSE", icon='line_chart_blue')
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        self.config_path = config_path
        self.settings_frm = LabelFrame(self.main_frm, text="SETTINGS", font=Formats.FONT_HEADER.value)
        self.type_dropdown = DropDownMenu(self.settings_frm, "INTERPOLATION TYPE:", ['MISSING BODY-PARTS', 'MISSING ANIMALS'], "35")
        self.method_dropdown = DropDownMenu(self.settings_frm, "INTERPOLATION METHOD:", ['NEAREST', 'LINEAR', 'QUADRATIC'], "35")
        self.save_originals_dropdown = DropDownMenu(self.settings_frm, "SAVE ORIGINALS:", Options.BOOL_STR_OPTIONS.value, "35")
        self.type_dropdown.setChoices('MISSING BODY-PARTS')
        self.method_dropdown.setChoices('NEAREST')
        self.save_originals_dropdown.setChoices(Options.BOOL_STR_OPTIONS.value[0])

        self.settings_frm.grid(row=0, column=0, sticky=NW)
        self.type_dropdown.grid(row=0, column=0, sticky=NW)
        self.method_dropdown.grid(row=1, column=0, sticky=NW)
        self.save_originals_dropdown.grid(row=2, column=0, sticky=NW)

        self.single_file_frm = LabelFrame(self.main_frm, text="INTERPOLATE SINGLE DATA FILE", font=Formats.FONT_HEADER.value)
        self.selected_file = FileSelect(self.single_file_frm, "DATA PATH:", lblwidth=35, file_types=[("VIDEO FILE", ".csv .parquet")], initialdir=self.project_path)
        self.run_btn_single = SimbaButton(parent=self.single_file_frm, txt="RUN SINGLE DATA FILE INTERPOLATION", img='rocket', txt_clr="blue", font=Formats.FONT_REGULAR.value, cmd=self.run, cmd_kwargs={'multiple': False})

        self.single_file_frm.grid(row=1, column=0, sticky=NW)
        self.selected_file.grid(row=0, column=0, sticky=NW)
        self.run_btn_single.grid(row=1, column=0, sticky=NW)

        self.multiple_file_frm = LabelFrame(self.main_frm, text="INTERPOLATE DIRECTORY OF DATA", font=Formats.FONT_HEADER.value)
        self.selected_dir = FolderSelect(self.multiple_file_frm, "SELECT DIRECTORY OF DATA FILES:", lblwidth=35, initialdir=self.project_path)


        self.run_btn_multiple = SimbaButton(parent=self.multiple_file_frm, txt="RUN DATA DIRECTORY INTERPOLATION", img='rocket', txt_clr="blue", font=Formats.FONT_REGULAR.value, cmd=self.run, cmd_kwargs={'multiple': True})
        self.multiple_file_frm.grid(row=2, column=0, sticky=NW)
        self.selected_dir.grid(row=0, column=0, sticky=NW)
        self.run_btn_multiple.grid(row=1, column=0, sticky=NW)
        self.main_frm.mainloop()

    def run(self, multiple):
        interpolation_type = INTERPOLATOR_METHOD[self.type_dropdown.getChoices()]
        interpolation_method = self.method_dropdown.getChoices().lower()
        copy_originals = str_2_bool(self.save_originals_dropdown.getChoices())
        if not multiple:
            data_path = self.selected_file.file_path
            check_file_exist_and_readable(file_path=data_path)
            data_dir = Path(os.path.dirname(data_path))
        else:
            data_path = self.selected_dir.folder_path
            check_if_dir_exists(in_dir=data_path)
            data_dir = Path(data_path)

        multi_index_df_headers = False
        if data_dir.resolve().absolute() == Path(self.input_csv_dir).resolve().absolute():
            multi_index_df_headers = True
        interpolator = Interpolate(config_path=self.config_path,
                                   data_path=data_path,
                                   type=interpolation_type,
                                   method=interpolation_method,
                                   copy_originals=copy_originals,
                                   multi_index_df_headers=multi_index_df_headers)
        interpolator.run()


#InterpolatePopUp(config_path=r"C:\troubleshooting\my_project_tutorial\project_folder\project_config.ini")


__author__ = "Simon Nilsson"

import os
from typing import Dict, List, Optional, Union

import cv2
import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log,
    check_file_exist_and_readable, check_instance, check_valid_lst)
from simba.utils.errors import (CountError, InvalidInputError,
                                NoSpecifiedOutputError)
from simba.utils.lookups import get_color_dict
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, read_df


class DistancePlotterSingleCore(ConfigReader):
    """
    Class for visualizing the distance between two pose-estimated body-parts (e.g., two animals) through line
    charts. Results are saved as individual line charts, and/or videos of line charts.

    .. note::
       For better runtime, use :meth:`simba.plotting.distance_plotter_mp.DistancePlotterMultiCore`.
       `GitHub tutorial/documentation <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#step-11-visualizations>`__.

    .. image:: _static/img/distance_plot.png
       :width: 300
       :align: center

    :parameter str config_path: path to SimBA project config file in Configparser format.
    :parameter bool frame_setting: If True, creates individual frames.
    :parameter bool video_setting: If True, creates videos

    :examples:

    >>> style_attr = {'width': 640, 'height': 480, 'line width': 6, 'font size': 8, 'opacity': 0.5}
    >>> line_attr = {0: ['Center_1', 'Center_2', 'Green'], 1: ['Ear_left_2', 'Ear_left_1', 'Red']}
    >>> distance_plotter = DistancePlotterSingleCore(config_path=r'MyProjectConfig', files_found=['test/two_c57s/project_folder/csv/outlier_corrected_movement_location/Video_1.csv'], frame_setting=False, video_setting=True, final_img=True)
    >>> distance_plotter.run()
    """

    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        data_paths: List[Union[str, os.PathLike]],
        style_attr: Dict[str, int],
        line_attr: List[List[str]],
        frame_setting: Optional[bool] = False,
        video_setting: Optional[bool] = False,
        final_img: Optional[bool] = False,
    ):

        if (not frame_setting) and (not video_setting) and (not final_img):
            raise NoSpecifiedOutputError(
                msg="Please choice to create frames and/or video distance plots",
                source=self.__class__.__name__,
            )
        check_instance(
            source=f"{self.__class__.__name__} line_attr",
            instance=line_attr,
            accepted_types=(list,),
        )
        for cnt, i in enumerate(line_attr):
            check_valid_lst(
                source=f"{self.__class__.__name__} line_attr {cnt}",
                data=i,
                valid_dtypes=(str,),
                exact_len=3,
            )
        check_valid_lst(data=data_paths, valid_dtypes=(str,), min_len=1)
        _ = [check_file_exist_and_readable(i) for i in data_paths]
        ConfigReader.__init__(self, config_path=config_path)
        (
            self.video_setting,
            self.frame_setting,
            self.data_paths,
            self.style_attr,
            self.line_attr,
            self.final_img,
        ) = (video_setting, frame_setting, data_paths, style_attr, line_attr, final_img)
        self.color_names = get_color_dict()

    def run(self):
        print(f"Processing {len(self.data_paths)} videos...")
        check_all_file_names_are_represented_in_video_log(
            video_info_df=self.video_info_df, data_paths=self.data_paths
        )
        for file_cnt, file_path in enumerate(self.data_paths):
            video_timer = SimbaTimer(start=True)
            data_df = read_df(file_path, self.file_type)
            _, video_name, _ = get_fn_ext(file_path)
            self.video_info, px_per_mm, fps = self.read_video_info(
                video_name=video_name
            )
            self.save_video_folder = os.path.join(self.line_plot_dir, video_name)
            self.save_frame_folder_dir = os.path.join(self.line_plot_dir, video_name)

            try:
                data_df.columns = self.bp_headers
            except ValueError:
                raise CountError(
                    msg=f"SimBA expects {self.bp_headers} columns but found {len(data_df)} columns in {file_path}",
                    source=self.__class__.__name__,
                )
            distances = []
            colors = []
            for cnt, i in enumerate(self.line_attr):
                if i[2] not in list(self.color_names.keys()):
                    raise InvalidInputError(
                        msg=f"{i[2]} is not a valid color. Options: {list(self.color_names.keys())}.",
                        source=self.__class__.__name__,
                    )
                colors.append(i[2])
                bp_1, bp_2 = [f"{i[0]}_x", f"{i[0]}_y"], [f"{i[1]}_x", f"{i[1]}_y"]
                if len(list(set(bp_1) - set(data_df.columns))) > 0:
                    raise InvalidInputError(
                        msg=f"Could not find fields {bp_1} in {file_path}",
                        source=self.__class__.__name__,
                    )
                if len(list(set(bp_2) - set(data_df.columns))) > 0:
                    raise InvalidInputError(
                        msg=f"Could not find fields {bp_2} in {file_path}",
                        source=self.__class__.__name__,
                    )
                distances.append(
                    FeatureExtractionMixin.framewise_euclidean_distance(
                        location_1=data_df[bp_1].values,
                        location_2=data_df[bp_2].values,
                        px_per_mm=px_per_mm,
                        centimeter=True,
                    )
                )
            if self.frame_setting:
                if os.path.exists(self.save_frame_folder_dir):
                    self.remove_a_folder(self.save_frame_folder_dir)
                os.makedirs(self.save_frame_folder_dir)
            if self.video_setting:
                save_video_path = os.path.join(self.line_plot_dir, f"{video_name}.avi")
                fourcc = cv2.VideoWriter_fourcc(*"DIVX")
                video_writer = cv2.VideoWriter(
                    save_video_path,
                    fourcc,
                    fps,
                    (self.style_attr["width"], self.style_attr["height"]),
                )

            if self.final_img:
                _ = PlottingMixin.make_line_plot(
                    data=distances,
                    colors=colors,
                    width=self.style_attr["width"],
                    height=self.style_attr["height"],
                    line_width=self.style_attr["line width"],
                    font_size=self.style_attr["font size"],
                    title="Animal distances",
                    y_lbl="distance (cm)",
                    x_lbl="time (s)",
                    x_lbl_divisor=fps,
                    y_max=self.style_attr["y_max"],
                    line_opacity=self.style_attr["opacity"],
                    save_path=os.path.join(
                        self.line_plot_dir, f"{video_name}_final_distances.png"
                    ),
                )

            if self.video_setting or self.frame_setting:
                if self.style_attr["y_max"] == -1:
                    self.style_attr["y_max"] = max([np.max(x) for x in distances])
                for frm_cnt in range(distances[0].shape[0]):
                    line_data = [x[:frm_cnt] for x in distances]
                    img = PlottingMixin.make_line_plot_plotly(
                        data=line_data,
                        colors=colors,
                        width=self.style_attr["width"],
                        height=self.style_attr["height"],
                        line_width=self.style_attr["line width"],
                        font_size=self.style_attr["font size"],
                        title="Animal distances",
                        y_lbl="distance (cm)",
                        x_lbl="frame count",
                        x_lbl_divisor=fps,
                        y_max=self.style_attr["y_max"],
                        line_opacity=self.style_attr["opacity"],
                        save_path=None,
                    ).astype(np.uint8)
                    if self.video_setting:
                        video_writer.write(img[:, :, :3])
                    if self.frame_setting:
                        frm_name = os.path.join(
                            self.save_frame_folder_dir, f"{frm_cnt}.png"
                        )
                        cv2.imwrite(frm_name, np.uint8(img))
                    print(f"Distance frame created: {frm_cnt}, Video: {video_name} ...")
                if self.video_setting:
                    video_writer.release()
                video_timer.stop_timer()
                stdout_success(
                    msg=f"Distance visualizations created for {video_name} saved at {self.line_plot_dir}",
                    elapsed_time=video_timer.elapsed_time_str,
                )
        self.timer.stop_timer()
        stdout_success(
            msg=f"Distance visualizations complete for {len(self.data_paths)} video(s)",
            elapsed_time=self.timer.elapsed_time_str,
        )


# style_attr = {'width': 640, 'height': 480, 'line width': 6, 'font size': 12, 'y_max': -1, 'opacity': 0.5}
# line_attr = [['Center_1', 'Center_2', 'Green'], ['Ear_left_2', 'Ear_right_2', 'Red']]
# test = DistancePlotterSingleCore(config_path=r'/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                  frame_setting=True,
#                                  video_setting=True,
#                                  style_attr=style_attr,
#                                  final_img=True,
#                                  data_paths=['/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/csv/outlier_corrected_movement_location/Together_1.csv'],
#                                  line_attr=line_attr)
# test.run()


#
# style_attr = {'width': 640,
#               'height': 480,
#               'line width': 6,
#               'font size': 8,
#               'y_max': 'auto',
#               'opacity': 0.9}
# line_attr = {0: ['Center_1', 'Center_2', 'Green'], 1: ['Ear_left_2', 'Ear_left_1', 'Red']}
#
# test = DistancePlotterSingleCore(config_path=r'/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                  frame_setting=False,
#                                  video_setting=True,
#                                  style_attr=style_attr,
#                                  final_img=True,
#                                  files_found=['/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/machine_results/Together_1.csv'],
#                                  line_attr=line_attr)
# test.create_distance_plot()

# style_attr = {'width': 640, 'height': 480, 'line width': 6, 'font size': 8}
# line_attr = {0: ['Termite_1_Head_1', 'Termite_1_Thorax_1', 'Dark-red']}

# test = DistancePlotterSingleCore(config_path=r'/Users/simon/Desktop/envs/troubleshooting/Termites_5/project_folder/project_config.ini',
#                                  frame_setting=False,
#                        video_setting=True,
#                        style_attr=style_attr,
#                        files_found=['/Users/simon/Desktop/envs/troubleshooting/Termites_5/project_folder/csv/outlier_corrected_movement_location/termites_1.csv'],
#                        line_attr=line_attr)
# test.create_distance_plot()



import os
import warnings
from typing import Optional, Union

from simba.mixins.config_reader import ConfigReader
from simba.mixins.image_mixin import ImageMixin
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.roi_tools.roi_ui_mixin import ROI_mixin
from simba.roi_tools.roi_utils import get_pose_for_roi_ui
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists)
from simba.utils.enums import ROI_SETTINGS
from simba.utils.errors import InvalidInputError
from simba.utils.read_write import (find_all_videos_in_directory, get_fn_ext,
                                    get_video_meta_data)

warnings.filterwarnings("ignore")


WINDOW_SIZE = (775, 900)
SHOW_GRID_OVERLAY = 'SHOW_GRID_OVERLAY'

class ROI_ui(ROI_mixin, ConfigReader):

    """
    Main antry-point for drawing ROIs on videos.

    ..note::
      See ROI tutorial on `GitHub <https://github.com/sgoldenlab/simba/blob/master/docs/roi_tutorial_new_2025.md>`_ or `ReadTheDocs https://simba-uw-tf-dev.readthedocs.io/en/latest/tutorials_rst/roi_tutorial_new_2025.html>`_

    :example:
    >>> ROI_ui(config_path=r"C:\troubleshooting\mouse_open_field\project_folder\project_config.ini", video_path=r"C:\troubleshooting\mouse_open_field\project_folder\videos\Video1.mp4")
    >>> ROI_ui(config_path=r"/mnt/c/troubleshooting/mouse_open_field/project_folder/project_config.ini", video_path=r"/mnt/c/troubleshooting/mouse_open_field/project_folder/videos/Video1.mp4")

    """
    def __init__(self,
                 video_path: Union[str, os.PathLike],
                 config_path: Optional[Union[str, os.PathLike]] = None,
                 roi_coordinates_path: Optional[Union[str, os.PathLike]] = None,
                 pose_path: Optional[Union[str, os.PathLike]] = None,
                 video_dir: Optional[Union[str, os.PathLike]] = None):

        check_file_exist_and_readable(file_path=video_path)
        if config_path is None:
            if (video_dir is None) and (roi_coordinates_path is None):
                raise InvalidInputError(msg='If not passing a config path, then pass roi_coordinates_path AND video_dir', source=self.__class__.__name__)
        if config_path is not None:
            check_file_exist_and_readable(file_path=config_path)
            ConfigReader.__init__(self, config_path=config_path, read_video_info=False, create_logger=False)
        if roi_coordinates_path is not None:
            self.roi_coordinates_path = roi_coordinates_path
            if check_if_dir_exists(in_dir=video_dir, raise_error=False): self.video_dir = video_dir
            else: self.video_dir = None
        self.video_meta = get_video_meta_data(video_path=video_path, fps_as_int=False)
        if pose_path is not None:
            pose_data = get_pose_for_roi_ui(pose_path=pose_path, video_path=video_path)
        else:
            pose_data = None
        self.video_ext = get_fn_ext(filepath=video_path)[2][1:]
        self.img, self.img_idx = ImageMixin.find_first_non_uniform_clr_frm(video_path=video_path)
        self.define_ui = PopUpMixin(title="REGION OF INTEREST (ROI) SETTINGS", size=WINDOW_SIZE, icon='shapes_small')
        ROI_mixin.__init__(self, video_path=video_path, config_path=config_path, img_idx=self.img_idx, main_frm=self.define_ui.root, roi_coordinates_path=self.roi_coordinates_path, pose_data=pose_data)
        if self.video_dir is not None:
            self.other_project_video_paths = find_all_videos_in_directory(directory=self.video_dir, as_dict=True).values()
            self.other_project_video_paths = [x for x in self.other_project_video_paths if x != video_path]
            self.other_project_video_names = [get_fn_ext(x)[1] for x in self.other_project_video_paths]
        else:
            self.other_project_video_names = []
        self.settings = {item.name: item.value for item in ROI_SETTINGS}
        self.settings[SHOW_GRID_OVERLAY] = 'FALSE'
        self.get_video_info_panel(parent_frame=self.define_ui.main_frm, row_idx=0)
        self.get_select_img_panel(parent_frame=self.define_ui.main_frm, row_idx=1)
        self.get_select_shape_type_panel(parent_frame=self.define_ui.main_frm, row_idx=2)
        self.get_shape_attr_panel(parent_frame=self.define_ui.main_frm, row_idx=3)
        self.get_shape_name_panel(parent_frame=self.define_ui.main_frm, row_idx=4)
        self.get_interact_panel(parent_frame=self.define_ui.main_frm, row_idx=5, top_level=self.define_ui.root)
        self.get_draw_panel(parent_frame=self.define_ui.main_frm, row_idx=6, top_level=self.define_ui.root)
        self.get_shapes_from_other_video_panel(parent_frame=self.define_ui.main_frm, row_idx=7)
        self.get_save_roi_panel(parent_frame=self.define_ui.main_frm, row_idx=8)
        self.get_status_bar_panel(parent_frame=self.define_ui.main_frm, row_idx=9)
        self.get_file_menu(root=self.define_ui.root)
        self.set_selected_shape_type(shape_type='rectangle')
        self.define_ui.root.protocol("WM_DELETE_WINDOW", self._close)
        self.define_ui.main_frm.mainloop()

    def _close(self):
        try:
            self.close_img()
        except:
            pass
        self.define_ui.root.destroy()
        self.define_ui.root.quit()

# ROI_ui(config_path=r"C:\troubleshooting\SDS_pre_post\project_folder\project_config.ini",
#        video_path=r"C:\troubleshooting\SDS_pre_post\project_folder\videos\SDI100 x ALR2 pre_d7.mp4")
#



# ROI_ui(config_path=None,
#        video_path=r"D:\ares\data\ant\ant.mp4",
#        roi_coordinates_path=r'D:\ares\data\ant\roi_coordinates.h5')

# ROI_ui(config_path="/Users/simon/Desktop/envs/simba/troubleshooting/spontenous_alternation/project_folder/project_config.ini",
#        video_path="/Users/simon/Desktop/envs/simba/troubleshooting/spontenous_alternation/project_folder/videos/F1 HAB.mp4")

# ROI_ui(config_path=None,
#        video_path=r'C:\troubleshooting\mouse_open_field\project_folder\videos\Video2.mp4',
#        roi_coordinates_path=r"C:\troubleshooting\mitra\project_folder\logs\measures\ROI_definitions.h5",
#        video_dir=r"C:\troubleshooting\mitra\test")


# ROI_ui(config_path=r"C:\troubleshooting\RAT_NOR\project_folder\project_config.ini",
#        video_path=r"C:\troubleshooting\RAT_NOR\project_folder\videos\03152021_NOB_IOT_8.mp4",
#        roi_coordinates_path=r"C:\troubleshooting\RAT_NOR\project_folder\logs\measures\ROI_definitions.h5",
#        pose_path=r"C:\troubleshooting\RAT_NOR\project_folder\csv\outlier_corrected_movement_location\03152021_NOB_IOT_8.csv")


# ROI_ui(config_path=r"C:\troubleshooting\cue_light\t1\project_folder\project_config.ini",
#        video_path=r"C:\troubleshooting\cue_light\t1\project_folder\videos\2025-05-21 16-10-06_cropped.mp4",
#        roi_coordinates_path=r"C:\troubleshooting\cue_light\t1\project_folder\logs\measures\ROI_definitions.h5")
#


# ROI_ui(config_path=r"C:\troubleshooting\cue_light\t1\project_folder\project_config.ini",
#        video_path=r"C:\troubleshooting\cue_light\t1\project_folder\videos\2025-05-21 16-10-06_cropped.mp4",
#        roi_coordinates_path=r"C:\troubleshooting\cue_light\t1\project_folder\logs\measures\ROI_definitions.h5")
#



# ROI_ui(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini",
#        video_path=r"C:\troubleshooting\mitra\project_folder\videos\501_MA142_Gi_CNO_0514.mp4")
#
# #
#ROI_ui(config_path=r'/Users/simon/Desktop/envs/simba/troubleshooting/RAT_NOR/project_folder/project_config.ini',
#       video_path=r"/Users/simon/Desktop/envs/simba/troubleshooting/RAT_NOR/project_folder/videos/2022-06-20_NOB_DOT_4.mp4",
#       pose_path=r'/Users/simon/Desktop/envs/simba/troubleshooting/RAT_NOR/project_folder/csv/outlier_corrected_movement_location/2022-06-20_NOB_DOT_4.csv')


# ROI_ui(config_path=r'/Users/simon/Desktop/envs/simba/troubleshooting/open_field_below/project_folder/project_config.ini',
#       video_path=r"/Users/simon/Desktop/envs/simba/troubleshooting/open_field_below/project_folder/videos/raw_clip1.mp4",
#       pose_path=r'/Users/simon/Desktop/envs/simba/troubleshooting/open_field_below/project_folder/csv/outlier_corrected_movement_location/raw_clip1.csv')


# ROI_ui(config_path=r"C:\troubleshooting\open_field_below\project_folder\project_config.ini",
#       video_path=r"C:\troubleshooting\open_field_below\project_folder\videos\raw_clip1.mp4",
#       pose_path=r"C:\troubleshooting\open_field_below\project_folder\csv\outlier_corrected_movement_location\raw_clip1.csv")
#
#
# ROI_ui(config_path=r"C:\troubleshooting\open_field_below\project_folder\project_config.ini",
#       video_path=r"C:\troubleshooting\open_field_below\project_folder\videos\raw_clip1.mp4")

# ROI_ui(config_path=r"C:\troubleshooting\two_black_animals_14bp\project_folder\project_config.ini",
#       video_path=r"C:\troubleshooting\two_black_animals_14bp\project_folder\videos\Together_1.avi")



__author__ = "Simon Nilsson"

import multiprocessing
import os
from tkinter import *
from typing import Union

import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.plotting.heat_mapper_location import HeatmapperLocationSingleCore
from simba.plotting.heat_mapper_location_mp import \
    HeatMapperLocationMultiprocess
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        SimbaButton, SimbaCheckbox,
                                        SimBADropDown)
from simba.utils.checks import check_if_filepath_list_is_empty
from simba.utils.enums import Formats, Keys, Links, Paths
from simba.utils.read_write import get_file_name_info_in_directory


class HeatmapLocationPopup(PopUpMixin, ConfigReader):

    """
    :example:
    >>> _ = HeatmapLocationPopup(config_path=r"C:\troubleshooting\RAT_NOR\project_folder\project_config.ini")
    >>> _ = HeatmapLocationPopup(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini")

    """
    def __init__(self,
                 config_path: Union[str, os.PathLike]):

        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        self.data_path = os.path.join(self.project_path, Paths.OUTLIER_CORRECTED.value)
        self.files_found_dict = get_file_name_info_in_directory(directory=self.data_path, file_type=self.file_type)
        check_if_filepath_list_is_empty(filepaths=list(self.files_found_dict.keys()),error_msg="SIMBA ERROR: Zero files found in the project_folder/csv/outlier_corrected_movement_location directory. ",)
        PopUpMixin.__init__(self, title="HEATMAPS: LOCATION", icon='heatmap')
        max_scales = list(np.arange(5, 105, 5))
        max_scales.insert(0, "Auto-compute")
        self.style_settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="STYLE SETTINGS", icon_name=Keys.DOCUMENTATION.value, icon_link=Links.HEATMAP_LOCATION.value)

        self.palette_dropdown = SimBADropDown(parent=self.style_settings_frm, dropdown_options=self.palette_options, label='PALETTE: ', label_width=25, dropdown_width=30, value=self.palette_options[0])
        self.shading_dropdown = SimBADropDown(parent=self.style_settings_frm, dropdown_options=self.shading_options, label='SHADING: ', label_width=25, dropdown_width=30, value=self.shading_options[0])
        self.bp_dropdown = SimBADropDown(parent=self.style_settings_frm, dropdown_options=self.body_parts_lst, label='BODY-PART: ', label_width=25, dropdown_width=30, value=self.body_parts_lst[0])
        self.max_time_scale_dropdown = SimBADropDown(parent=self.style_settings_frm, dropdown_options=max_scales, label='MAX TIME SCALE (S): ', label_width=25, dropdown_width=30, value=max_scales[0])
        self.bin_size_dropdown = SimBADropDown(parent=self.style_settings_frm, dropdown_options=self.heatmap_bin_size_options, label='BIN SIZE (MM): ', label_width=25, dropdown_width=30, value="80×80")

        self.settings_frm = LabelFrame(self.main_frm, text="VISUALIZATION SETTINGS", font=Formats.FONT_HEADER.value, pady=5, padx=5)
        self.multiprocess_dropdown = SimBADropDown(parent=self.settings_frm, dropdown_options=list(range(1, self.cpu_cnt)), label='CPU CORES: ', label_width=25, dropdown_width=30, value=int(self.cpu_cnt/2))
        heatmap_frames_cb, self.heatmap_frames_var = SimbaCheckbox(parent=self.settings_frm, txt='CREATE FRAMES', txt_img='frames')
        heatmap_videos_cb, self.heatmap_videos_var = SimbaCheckbox(parent=self.settings_frm, txt='CREATE VIDEOS', txt_img='video')
        heatmap_last_frm_cb, self.heatmap_last_frm_var = SimbaCheckbox(parent=self.settings_frm, txt='CREATE LAST FRAME', txt_img='finish')

        self.run_frm = LabelFrame(self.main_frm, text="RUN", font=Formats.FONT_HEADER.value, pady=5, padx=5, fg="black")
        self.run_single_video_frm = LabelFrame(self.run_frm, text="SINGLE VIDEO", font=Formats.FONT_HEADER.value, pady=5, padx=5, fg="black")


        self.run_single_video_btn = SimbaButton(parent=self.run_single_video_frm, txt="Create single video", txt_clr="blue", img='rocket', font=Formats.FONT_REGULAR.value, cmd=self.__create_heatmap_plots, cmd_kwargs={'multiple_videos': False})
        self.single_video_dropdown = DropDownMenu(self.run_single_video_frm, "Video:", list(self.files_found_dict.keys()), "12")
        self.single_video_dropdown.setChoices(list(self.files_found_dict.keys())[0])
        self.run_multiple_videos = LabelFrame(self.run_frm, text="MULTIPLE VIDEO", font=Formats.FONT_HEADER.value, pady=5, padx=5, fg="black")

        self.run_multiple_video_btn = SimbaButton(parent=self.run_multiple_videos, txt="Create multiple videos ({} video(s) found)".format(str(len(list(self.files_found_dict.keys())))), img='rocket', txt_clr="blue", font=Formats.FONT_REGULAR.value, cmd=self.__create_heatmap_plots, cmd_kwargs={'multiple_videos': True})
        self.style_settings_frm.grid(row=0, sticky=NW)
        self.palette_dropdown.grid(row=0, sticky=NW)
        self.shading_dropdown.grid(row=1, sticky=NW)
        self.bp_dropdown.grid(row=2, sticky=NW)
        self.max_time_scale_dropdown.grid(row=3, sticky=NW)
        self.bin_size_dropdown.grid(row=4, sticky=NW)

        self.settings_frm.grid(row=1, sticky=NW)
        heatmap_frames_cb.grid(row=0, sticky=NW)
        heatmap_videos_cb.grid(row=1, sticky=NW)
        heatmap_last_frm_cb.grid(row=2, sticky=NW)
        self.multiprocess_dropdown.grid(row=3, column=0, sticky=NW)

        self.run_frm.grid(row=2, sticky=NW)
        self.run_single_video_frm.grid(row=0, sticky=NW)
        self.run_single_video_btn.grid(row=0, column=0, sticky=NW)
        self.single_video_dropdown.grid(row=0, column=1, sticky=NW)

        self.run_multiple_videos.grid(row=1, sticky=NW)
        self.run_multiple_video_btn.grid(row=0, sticky=NW)
        #self.main_frm.mainloop()

    def __create_heatmap_plots(self, multiple_videos: bool):
        if multiple_videos:
            data_paths = list(self.files_found_dict.values())
        else:
            data_paths = [self.files_found_dict[self.single_video_dropdown.getChoices()]]

        if self.max_time_scale_dropdown.getChoices() != "Auto-compute":
            max_scale = int(float(self.max_time_scale_dropdown.getChoices()))
        else:
            max_scale = "auto"

        bin_size = int(self.bin_size_dropdown.getChoices().split("×")[0])

        style_attr = {"palette": self.palette_dropdown.getChoices(),
                      "shading": self.shading_dropdown.getChoices(),
                      "max_scale": max_scale,
                      "bin_size": bin_size}

        if int(self.multiprocess_dropdown.get_value()) == 1:
            heatmapper_clf = HeatmapperLocationSingleCore(config_path=self.config_path,
                                                          style_attr=style_attr,
                                                          final_img_setting=self.heatmap_last_frm_var.get(),
                                                          video_setting=self.heatmap_videos_var.get(),
                                                          frame_setting=self.heatmap_frames_var.get(),
                                                          bodypart=self.bp_dropdown.getChoices(),
                                                          data_paths=data_paths)

            heatmapper_clf.run()

        else:
            heatmapper_clf = HeatMapperLocationMultiprocess(config_path=self.config_path,
                                                            style_attr=style_attr,
                                                            final_img_setting=self.heatmap_last_frm_var.get(),
                                                            video_setting=self.heatmap_videos_var.get(),
                                                            frame_setting=self.heatmap_frames_var.get(),
                                                            bodypart=self.bp_dropdown.getChoices(),
                                                            data_paths=data_paths,
                                                            core_cnt=int(self.multiprocess_dropdown.getChoices()))

            heatmapper_clf.run()
